from .download import download_zip, download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_zip(xz_url + "hair_baking.zip",down_path)
download_zip(xz_url + "hair_library.zip",down_path)
download_file(xz_url + "hair_lib.blend",down_path)
download_file(xz_url + "shell_hair_geo_nodes.blend",down_path)
download_file(xz_url + "shell_hair_texture_bake.blend",down_path)
download_file(xz_url + "short_hair_geo_nodes.blend",down_path)


bl_info = {
    "name": "Hair Tool",
    "description": "Procedural Hair Modeling Tools",
    "author": "Bartosz Styperek",
    "blender": (3, 6, 0),
    "location": "View 3D > Tools > Hair Tool",
    "version": (3, 6, 6),
    "warning": "",
    "doc_url": "https://joseconseco.github.io/HairTool_3_Documentation/",
    "tracker_url": "https://discord.gg/cxZDbqH",
    "category": "Object"
}
new_bl_info = bl_info

# load and reload submodules
##################################

if "bpy" in locals():
    import importlib

    importlib.reload(hair_tool_ui)
    importlib.reload(gpu_helpers)
    importlib.reload(curve_wrapper)
    importlib.reload(general_utils)
    importlib.reload(mesh_utils)
    importlib.reload(helper_functions)
    importlib.reload(finalize_to_copy)
    importlib.reload(hair_common)
    importlib.reload(hair_geometry_nodes_profile)
    # importlib.reload(generate_braids)
    importlib.reload(curves_from_grid)
    importlib.reload(ribbons_operations)
    importlib.reload(material_operators)
    importlib.reload(resample2d)
    importlib.reload(hair_curve_helpers)
    importlib.reload(hair_mesh_helpers)
    importlib.reload(curve_simplify)
    importlib.reload(short_hair)
    importlib.reload(shell_hair)
    importlib.reload(uv_draw_boxes)
    importlib.reload(curves_resample)
    importlib.reload(drawing_curves)
    importlib.reload(hair_tool_props)
    importlib.reload(hair_bake)
    importlib.reload(hair_bake_ui)
    importlib.reload(hair_bake_nodes)
    importlib.reload(ht_image_operators)
    importlib.reload(hair_geometry_nodes_shared)
    importlib.reload(hair_geometry_nodes_utils)
    importlib.reload(hair_geometry_nodes_selections)
    importlib.reload(hair_geometry_nodes_ui)
    importlib.reload(hair_geometry_nodes_uvs)
    importlib.reload(hair_geometry_nodes)
    importlib.reload(hair_geometry_nodes_drawing)
    importlib.reload(hair_geometry_nodes_update)
    importlib.reload(hair_geometry_nodes_shape_keys)
    importlib.reload(curves_attributes_manager)
    importlib.reload(mi_utils_base)
    importlib.reload(mi_widget_linear_deform)
    importlib.reload(deformer_curve)
    importlib.reload(generate_curls)
    importlib.reload(hair_workspace_tool)
    importlib.reload(hair_lib)
    importlib.reload(modal_save_hair)
    importlib.reload(jiggle_chain)
    importlib.reload(jiggle_chain_b4)
    importlib.reload(particle_hair)
else:
    from . import hair_tool_ui
    from .utils import gpu_helpers
    from .utils import curve_wrapper
    from .utils import general_utils
    from .utils import mesh_utils
    from .utils import helper_functions
    from .utils import finalize_to_copy
    from .utils import hair_common
    # from . import generate_braids
    from . import curves_from_grid
    from . import material_operators
    from . import ribbons_operations
    from . import resample2d
    from . import hair_curve_helpers
    from . import hair_mesh_helpers
    from . import curve_simplify
    from . import short_hair
    from . import shell_hair
    from . import uv_draw_boxes
    from . import curves_resample
    from . import drawing_curves
    from . import hair_tool_props
    from .hair_baking import hair_geometry_nodes_profile
    from .hair_baking import hair_bake
    from .hair_baking import hair_bake_ui
    from .hair_baking import hair_bake_nodes
    from .hair_baking import ht_image_operators
    from .hair_baking import hair_geometry_nodes_shared
    from .hair_baking import hair_geometry_nodes_utils
    from .hair_baking import hair_geometry_nodes_selections
    from .hair_baking import hair_geometry_nodes_ui
    from .hair_baking import hair_geometry_nodes_uvs
    from .hair_baking import hair_geometry_nodes
    from .hair_baking import hair_geometry_nodes_drawing
    from .hair_baking import hair_geometry_nodes_update
    from .hair_baking import hair_geometry_nodes_shape_keys
    from .hair_baking import curves_attributes_manager
    from .linear_deformer import mi_utils_base
    from .linear_deformer import mi_widget_linear_deform
    from .linear_deformer import deformer_curve
    from . import generate_curls
    from . import hair_workspace_tool

    from .hair_library import hair_lib
    from .hair_library import modal_save_hair
    from . import jiggle_chain
    from . import jiggle_chain_b4
    from . import particle_hair


# register
##################################
import bpy
from bpy.app.handlers import persistent
from . import auto_load
from .drawing_curves import change_draw_keymap
# from .hair_workspace_tool import GiGroup
# from .hair_workspace_tool import HTOOL_OT_ToolModelHair
auto_load.init()


def draw_curve_sel(self, context):
    layout = self.layout
    layout.separator()
    layout.operator_context = "INVOKE_DEFAULT"
    op = layout.operator("curves.ht_select_next", text="Select Next")
    op.previous = False
    op = layout.operator("curves.ht_select_next", text="Select Previous")
    op.previous = True
    layout.separator()
    layout.operator("curves.select_tips", text="Select Roots").roots = True
    layout.operator("curves.select_tips", text="Select Tips").roots = False

def draw_curve_sel_obj(self, context):
    called_from_ht_wspace = context.workspace.tools.from_space_view3d_mode(mode=context.mode).idname == "hair_tool.hair_transform"
    obj = context.active_object
    if called_from_ht_wspace and obj and obj.type == 'CURVES':
        layout = self.layout
        layout.separator()
        layout.operator_context = "INVOKE_DEFAULT"
        layout.operator("curves.select_tips")


def draw_save_file_node(self, context):
    layout = self.layout
    node = context.active_node
    if context.active_node and context.active_node.type == 'OUTPUT_FILE':
        layout.operator("node.save_image_file_node")
        layout.separator()


def draw_curves_overlay(self, context):
    # if context.active_object and context.active_object.type  == 'CURVES':
    # box = self.layout.box()
    layout = self.layout
    ht_scene_props = context.scene.ht_props
    col = layout.column()
    col.label(text='Curves Overlay')
    split = col.split(factor=0.8)
    split.prop(ht_scene_props,'strands_overlay', text='Draw Strands Overlay') # , icon="OVERLAY", toggle=True
    # row.popover(panel="HT_PT_CurvesDrawPopup", icon='DOWNARROW_HLT', text='')
    if ht_scene_props.strands_overlay:
        addon_prefs = general_utils.get_addon_preferences()
        split.prop(addon_prefs,'draw_color', text='')
        row = col.row(align=True)
        row.prop(addon_prefs,'draw_offset')




@persistent
def post_load(scene):
    #Scene is None, when addosn is initiailzed. And Keymap'Grease Pencilt' it none too. Lines below work only on new blend load. :<
    use_custom_hair_drawing = bpy.context.scene.ht_props.hair_drawing.draw_strands
    change_draw_keymap(use_custom_hair_drawing)
    bpy.context.scene.ht_props.jiggle_props.sim_is_running = False
    bpy.context.scene.ht_props.strands_overlay = False # disable hair system guide drawing


def register():
    # bpy.utils.register_class(hair_workspace_tool.GiGroup)
    auto_load.register()
    bpy.utils.register_tool(hair_workspace_tool.HTOOL_OT_ToolModelHair, after={"builtin.scale_cage"}, separator=True, group=True)
    bpy.utils.register_tool(hair_workspace_tool.HTOOL_OT_ToolModelHairSculptMode, after={"builtin_brush.Slide"}, separator=True, group=True)

    print("Registered Hair Tool")

    addon_prefs = general_utils.get_addon_preferences()
    addon_prefs.update_text = ''

    hair_tool_props.register_properties()
    hair_bake_nodes.registerNode()

    # lets add ourselves to the main header
    bpy.types.VIEW3D_MT_select_edit_curves.append(draw_curve_sel)
    bpy.types.VIEW3D_MT_select_object.append(draw_curve_sel_obj)
    bpy.types.NODE_PT_active_node_properties.prepend(draw_save_file_node)

    bpy.types.VIEW3D_PT_overlay.append(draw_curves_overlay)


    # key_map_prefs.register_keymap()
    # dealay register_keymap so that it sees wm.keyconfigs.default - and can properly override it
    # bpy.app.timers.register(hair_tool_ui.register_keymap, first_interval=1)
    # WARN: somehow if loading blend file, timer delayed register hack wont work... Use sleep instead
    import time
    time.sleep(.1)
    hair_tool_ui.register_keymap()

    # hair_tool_ui.update_panel(None, bpy.context) # just complicates thing

    bpy.app.handlers.load_post.append(post_load)
    bpy.types.Scene.last_stroke = {}

    hair_lib.reg()
    uv_draw_boxes.append_IMAGE_HT_header()


def unregister():
    bpy.context.scene.ht_props.strands_overlay = False
    uv_draw_boxes.remove_IMAGE_HT_header()
    hair_lib.unreg()

    bpy.app.handlers.load_post.remove(post_load)
    hair_bake_nodes.unregisterNode()
    bpy.context.scene.ht_props.hair_drawing.draw_strands = False  # disable hair drawing, since it wont work on F8 anyway
    hair_tool_ui.unregister_keymap()
    hair_tool_props.unregister_properties()

    bpy.utils.unregister_tool(hair_workspace_tool.HTOOL_OT_ToolModelHair)
    bpy.utils.unregister_tool(hair_workspace_tool.HTOOL_OT_ToolModelHairSculptMode)
    # bpy.utils.unregister_class(hair_workspace_tool.GiGroup)
    auto_load.unregister()
    del bpy.types.Scene.last_stroke
    print("Unregistered Hair Tool.")

    bpy.types.VIEW3D_PT_overlay.remove(draw_curves_overlay)
    bpy.types.VIEW3D_MT_select_edit_curves.remove(draw_curve_sel)
    bpy.types.NODE_PT_active_node_properties.remove(draw_save_file_node)
    bpy.types.VIEW3D_MT_select_object.remove(draw_curve_sel_obj)


if __name__ == "__main__":
    register()
