# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO
# import sys
# dir = 'C:\\Users\\JoseConseco\\AppData\\Local\\Programs\\Python\\Python36\\Lib\\site-packages'
# if not dir in sys.path:
#     sys.path.append(dir )
# import ipdb

import bpy
from math import log
from bl_math import lerp
import numpy as np
from bpy.props import StringProperty, IntProperty, FloatProperty, BoolProperty
from .hair_baking.hair_bake import setup_pass, ht_channel_mixing
from .hair_baking.hair_geometry_nodes import change_active_mod, set_tag_from_prop_update, set_parent_from_prop_update, \
    set_filter_from_prop_update
from .hair_baking.hair_geometry_nodes_drawing import draw_new_hair, on_draw_guide_overlay_depsgraph_update
from .hair_baking.hair_geometry_nodes_selections import get_any_type_fselection, get_create_attrib_cache

from .short_hair import update_all_uv_new
from .drawing_curves import ModalDrawingSettings
from .hair_library.hair_lib import enum_previews_from_directory_items
from .utils.general_utils import get_addon_preferences
from .hair_library import hair_lib
import os
from .hair_baking.hair_geometry_nodes_shared import get_profile_mod, uv_owner_input_name, auto_update_all_curves_uv_regions

def set_bone_col(pbone, color): # from jiggle_chain_b4..., only for blender 4. and up
    # # Set theme color:
    # bone.color.palette = 'THEME09'
    pbone.bone.color.palette = 'CUSTOM'
    pbone.bone.color.custom.normal = color
    pbone.bone.color.custom.select = (1.0, 0.9, 0.3)
    pbone.bone.color.custom.active = (1.0, 1.0, 1.0)
    pbone.color.palette = 'CUSTOM'
    pbone.color.custom.normal = color
    pbone.color.custom.select = (1.0, 0.9, 0.3)
    pbone.color.custom.active = (1.0, 1.0, 1.0)


class hairUVPoints(bpy.types.PropertyGroup):
    start_point: bpy.props.FloatVectorProperty(name="", description="", default=(0.0, 0.0), size=2)
    end_point: bpy.props.FloatVectorProperty(name="", description="", default=(1.0, 1.0), size=2)
    offset_tip: FloatProperty(name="Offset to tip", description="Move uv's toward tip", default=0, min=0, max=1, subtype='PERCENTAGE')
    offset_root: FloatProperty(name="Offset to root", description="Move uv's toward root", default=0, min=0, max=1, subtype='PERCENTAGE')
    flip_y: bpy.props.BoolProperty(name='flip y', description='Flip uv hair root to tip direction', default=False)



UV_REGIONS_IDS = []
def uv_box_items(self, context):
    active_obj = context.active_object
    profile_mod = get_profile_mod(active_obj)
    global UV_REGIONS_IDS
    uv_owner = profile_mod[uv_owner_input_name]
    box_cnt = len(uv_owner.data.vertices)
    # uv_start_attr = uv_owner.data.attributes.get("UV_Start")
    # all_regions = ('ALL','All', 'Use All UV regions randomly')
    if box_cnt != len(UV_REGIONS_IDS): # +1 for all reg
        UV_REGIONS_IDS = [(str(i), ''+str(i), '') for i in range(box_cnt)]
    return UV_REGIONS_IDS



class hair_ribbons_props(bpy.types.PropertyGroup):
    hair_mesh_parent: StringProperty(name="", default="")
    hair_curve_child: StringProperty(name="", default="")

class ObjShortHairProps(bpy.types.PropertyGroup):
    is_short_hair: bpy.props.BoolProperty(name='Is short hair Mesh', description='Was Created As Short hair mesh', default=False)
    # is_short_hair_output: bpy.props.IntProperty(name='Is short hair Output mesh', description='Was Created As Short hair mesh', default=0, min=0, max=1)
    hair_base_mesh: StringProperty(name="", default="")

class ObjShellHairProps(bpy.types.PropertyGroup):
    # replaced by: get_shell_hair_mod(obj)
    # is_shell_hair: bpy.props.BoolProperty(name='Is short hair Mesh', description='Was Created As Short hair mesh', default=False)
    # is_short_hair_output: bpy.props.IntProperty(name='Is short hair Output mesh', description='Was Created As Short hair mesh', default=0, min=0, max=1)
    hair_base_mesh: StringProperty(name="", default="")

class ObjBoneGenProps(bpy.types.PropertyGroup):
    absolete_start: bpy.props.FloatProperty(name="Absolute Start", description='Absolute offset from strand root location', default=0.0, min=0.0, soft_max=10.0, subtype='DISTANCE')
    relative_start: bpy.props.IntProperty(name="Chain relative start", description='Place first bone at some start distance from strand root location', default=60, min=0, max=90, subtype='PERCENTAGE')
    to_tip_bias: bpy.props.IntProperty(name="Bias toward Tip", description='Clump generated bones more toward strand tip', default=0, min=0, max=100, subtype='PERCENTAGE')
    assign_jiggle: BoolProperty(name="Assign Jiggle", description="Assign jiggle to bones", default=True)
    generation_mode: bpy.props.EnumProperty(name='Bone generation mode', description='Descibes how new bones will be generated',
        items=[
            ('SELECTED_ISLANDS', 'For all selected stripes', 'Generate bone chain for all selected strans (mesh islands)'),
            ('ACTIVE_ISLAND', 'For active stripe only', 'Generate bone chain only for active stripe (island with active face/edge/vert)')
        ], default='ACTIVE_ISLAND')
    normalize_by_length: bpy.props.BoolProperty(name="Normalize by length", default=False, description="Bones count will depend on haircard length - longer strands will receive more bones in chain - up to 'Max Bones Count' for longest strand")
    bone_count: bpy.props.IntProperty(name="Bones count", description='', default=2, min=1, soft_max=8)
    max_bone_count: bpy.props.IntProperty(name="Max bones count", description='Maximum bones count for longest strand in normalized algorithm version', default=6, min=1, soft_max=16)


passes = (
    ("AO", "Ambient Occlusion", "Ambient Occlusion"),
    ("DIFFUSE", "Diffuse", "Diffuse mask"),
    ("OPACITY", "Opacity", "Opacity mask"),
    ("NORMAL", "Normal map", "Normal map"),
    ("ROOT", "Root map", "Gradient from hair root to tip"),
    ("ID", "Random ID", "Random grayscale value per hair strands"),
    ("RANDOM_PARENT", "Random Parent", "Random color per clumps of hair strands (based on parent strands ID)"),
    ("COLOR", "Random Color", "Random color per hair strands"),
    ("FLOW", "Flow map", "Drives direction of anisotropic specular (alternative to Directional map)"),
    ("DIRECTIONAL", "Directional map", "Alternative to Flow map - drives direction of anisotropic specular"),
    ("DEPTH", "Depth map", "Depth map"))


class HairBakeSettings(bpy.types.PropertyGroup):
    def update_materials(self, context):
        if len(self.render_passes)>0:
            setup_pass(context, self.render_passes.pop())

    def update_compose(self, context):
        if len(self.render_passes) > 0 and self.hair_bake_composite:
            ht_channel_mixing(context)


    def update_pad(self, context):
        if self.padding_mode == 'AUTO':
            self['padding_size'] = int(int(self.bakeResolution_x)/64)

    def update_ratio(self, context):
        if self.lock_ratio:
            self.bakeResolution_y = self.bakeResolution_x

    def update_res_x_ratio(self, context):
        if self.lock_ratio:
            self.bakeResolution_y = self.bakeResolution_x
        self.update_res_common(context)

    def update_res_common(self, context):
        render = context.scene.render
        render.resolution_x = int(self.bakeResolution_x)
        render.resolution_y = int(self.bakeResolution_y)
        cam = bpy.data.objects['Camera_render']
        x_to_y = render.resolution_x/render.resolution_y
        y_to_x = 1 / x_to_y
        def cam_shift_calc(div_ratio):
            c = int(log(div_ratio, 2)) # 2 ^y = div_ratio
            exponents = np.array(range(int(c)))+2
            return np.sum(np.power(0.5, exponents))
        cam.data.ortho_scale = max(pow(2, log(x_to_y*2, 2)), pow(2, log(y_to_x*2, 2)))
        # cam.data.shift_x = -1*cam_shift_calc(y_to_x) if y_to_x > 1 else 0
        # cam.data.shift_y = cam_shift_calc(x_to_y) if x_to_y > 1 else 0
        self.update_pad(context)


    render_quality: bpy.props.EnumProperty(name="Quality", default="2", description='Multiplier: [16 base samples] * [quality] -(1)Low, (2)Normal, (4)High, (8)Ultra',
                                           items=(
                                               ("1", "Low (16)", ""),
                                               ("2", "Normal (32)", ""),
                                               ("4", "High (64)", ""),
                                               ("8", "Ultra (128)", ""),
                                               ("16","Ludicrous (256)", ""),
                                           ))
    lock_ratio: bpy.props.BoolProperty(name='Lock Ratio', description='Keep X and Y resolution the same', default=True, update=update_ratio)
    bakeResolution_x: bpy.props.EnumProperty(name="Resolution X", default="1024",
                                            items=(("256",  "256", ""),
                                                   ("512",  "512", ""),
                                                   ("1024", "1024", ""),
                                                   ("2048", "2048", ""),
                                                   ("4096", "4096", ""),
                                                   ("8192", "8192", "")),
                                             update=update_res_x_ratio)
    bakeResolution_y: bpy.props.EnumProperty(name="Resolution Y", default="1024",
                                            items=(("256",  "256", ""),
                                                   ("512",  "512", ""),
                                                   ("1024", "1024", ""),
                                                   ("2048", "2048", ""),
                                                   ("4096", "4096", ""),
                                                   ("8192", "8192", "")),
                                             update=update_res_common)
    render_passes: bpy.props.EnumProperty(name="Pass", default={"NORMAL"},
                                           items=passes,
                                           options={'ENUM_FLAG'},
                                           update=update_materials)

    padding_mode: bpy.props.EnumProperty(name='Padding Mode', description='Prevents background color beeding into texture',
                                               items=[
                                                   ('AUTO', 'Automatic Padding', 'Padding will be set to around 1% of texture size'),
                                                   ('FIXED', 'Fixed Padding', 'Set padding size by hand')
                                               ], default='AUTO', update=update_pad)
    padding_size: bpy.props.IntProperty(name='Padding', description='Add color margin around bake result (0 - disabled)\nWarning - may be slow on big images', default = 10, min=0, soft_max=40)
    preview_background: bpy.props.EnumProperty(name='Background', description='Background for bake preview only',
        items=[
            ('TRANSPARENT', 'Transparent Background', 'Transparent'),
            ('SOLID', 'Solid Background', 'Solid')
        ], default='SOLID', update=update_materials)


    hair_bake_path: StringProperty(name="Path", default="", description="Baking output path", subtype='DIR_PATH')
    hair_bake_file_name: StringProperty(name="Name", default="Hair", description="Output texture name prefix")

    hair_bake_composite: BoolProperty(name="Mix Baked Channels", description="Mix RGBA channels of baked textures in HTool 'Texture Channel Mixing' nodes editor", default=False, update=update_compose)

    my_items = (
        ('PNG', "PNG", ""),
        ('TARGA', "TGA", ""),
        ('JPEG', "JPEG", "")
    )
    output_format: bpy.props.EnumProperty(name="Format", description="Format", items=my_items, default='PNG')


class curly_hair_settings(bpy.types.PropertyGroup):
    was_created_from_curls: BoolProperty(name="Created from Curls", description="Was curls operator used on current curve?", default=False)
    hairType: bpy.props.EnumProperty(name="Output Curve Type", default="NURBS",
                                     items=(("BEZIER", "Bezier", ""),
                                            ("NURBS", "Nurbs", ""),
                                            ("POLY", "Poly", "")))
    points_per_cycle: IntProperty(name="Points per cycle", description="How many to generate points per curl. Bigger values gives smoother curls", default=4, min=2, soft_max=10)
    curl_freq: FloatProperty(name="Curl frequency", description='Will generate more curls but also higher point count', default=5, min=0.4, soft_max=15)

    contrast: FloatProperty(name="Transition Contrast", description="Curls transition contrast", default=0.5, min=0, soft_max=1)
    offset: FloatProperty(name="Offset", description="Offset curls more toward the root or tip", default=0, min=-1, max=1)

    radius: FloatProperty(name="Radius", description="Curls Radius", default=1, min=0, soft_max=10)

    direction: bpy.props.EnumProperty(name='Direction', description='Curls Direction',
                                      items=[
                                          ('CW', 'Clockwise', ''),
                                          ('CCW', 'Counterclockwise', '')
                                      ], default='CW')

    random_direction: BoolProperty(name="Random direction", description="Flip curls direction randomly from clockwise to canterclockwise", default=False)

    gravity_tension: FloatProperty(name="Gravity ", description="Gravity weight pulls the curls down along the hair strand. O - disabled, 0.5 - natural gravity pull",
                                   default=0, min=0, max=1, subtype='PERCENTAGE')
    align_method: bpy.props.EnumProperty(name='Tilt align method', description='Tilt align method',
                                         items=[('ZUP', 'Z-up', 'Good of you have hair going down'),
                                                ('MINIMUM', 'Minimum', 'Not great but better than nothing')], default='MINIMUM')

    align_tilt: BoolProperty(name="Align Tilt", description="Make curve profile face outwards direction.\nIt is slow - usually it is best to enable it at last step", default=False)
    radomize_freq: FloatProperty(name="Randomize frequency", description="Randomize frequency", default=0, min=0, max=1, subtype='PERCENTAGE')
    # fix_angle: FloatProperty(name="fix_angle", description="Randomize frequency", default=0, min=-3.2, max=3.2)
    Seed: IntProperty(name="Noise Seed", default=1, min=1, max=500)

    uniformPointSpacing: BoolProperty(name="Uniform spacing", description="Distribute stand points with uniform spacing", default=True)
    # onlySelection: BoolProperty(name="Only Selected", description="Affect only selected splines", default=False)


class rand_tilt_settings(bpy.types.PropertyGroup):
    reset_first: BoolProperty(name="Reset on init", description="Clear tilt before randimizing it", default=True)
    # onlySelection: BoolProperty(name="Only Selected", description="Affect only selected points", default=True)
    tilt: FloatProperty(name="Tilt Angle", description="Add unifrom tilt to strands", default=1.5, min=-100, max=100, step=145, subtype='ANGLE')
    min_tilt: FloatProperty(name="Randomize", description="Tilt randomization strength", default=0.5, min=0, max=1, subtype='PERCENTAGE')
    tilt_transition_fallof: FloatProperty(name="Tilt falloff", description="Change braid tilt over strand length", default=0, min=0, max=1, subtype='PERCENTAGE')
    root_influence: FloatProperty(name="Root Influence", description="Make the hair tilt uniform across whole strand length", default=0, min=0, max=1, subtype='PERCENTAGE')
    randomDirection: BoolProperty(name="Flip Direction", description="When false hair will twist only clockwise", default=False)
    Seed: IntProperty(name="Noise Seed", default=1, min=1, max=500)


class radius_from_length_settings(bpy.types.PropertyGroup):
    min_radius: bpy.props.FloatProperty(name='Min Radius', description='', default= 0.5, min=0, max=1.0)
    max_radius: bpy.props.FloatProperty(name='Max Radius', description='', default= 1.0, min=0, max=1.0)
    len_threshold: bpy.props.FloatProperty(name='Transition threshold', description='Radius change will affect only strands close to given length threshold', default= 0.5, min=0, max=1)
    margin: bpy.props.FloatProperty(name='Transition Smoothness', description='Smooth out affected strand radius chagne transition', default=1.0, min=0, max=1.0)


class DeformersProps(bpy.types.PropertyGroup):
    '''Holds no data - just points to deformer node, by its index'''
    pass

class HairSystemProps(bpy.types.PropertyGroup):
    # def update_child(self, context):
        # set_as_child_mod(self)

    def update_tag(self, context):
        set_tag_from_prop_update(self, context)

    def update_filter(self, context):
        set_filter_from_prop_update(self, context)

    def update_parent_tag(self, context):
        set_parent_from_prop_update(self, context)

    deformers: bpy.props.CollectionProperty(type=DeformersProps)
    deformer_index: bpy.props.IntProperty(name="Active index", description='', default=1, min=0, max=100 )
    # DONE: handle when hsys is linked to mod with multi users...
    # set_as_child: bpy.props.BoolProperty(name="Set as child", description='Use strands from previous Hair System as target for some operations\n(useful for eg. clumping and twisting of hair strands)', default=False, update=update_child)

    # Helpers for [TAG], [PARENT] placeholders
    no_tag: bpy.props.StringProperty(name='No Tag', description='Tag to be used for strands that should not be affected by this system', default='[TAG]', update=update_tag)
    no_filter_tag: bpy.props.StringProperty(name='No Filter Tag', description='Tag to be used for strands that should not be affected by this system', default='[TAG]', update=update_filter)
    no_parent_tag: bpy.props.StringProperty(name='No Parent', description='Parent Tag to be used for current strands interpolation, clumping etc', default='[PARENT]', update=update_parent_tag)


class HairNodesProps(bpy.types.PropertyGroup):
    def idx_change(self, context):
        change_active_mod(self, context)

    hair_systems: bpy.props.CollectionProperty(type=HairSystemProps)
    system_index: bpy.props.IntProperty(name="Active index", description='', default=-1, min=0, max=100, update=idx_change)
    is_using_gnodes: bpy.props.IntProperty(name='Is using gnodes', description='', default=0)
    cyclic_dep_error: bpy.props.BoolProperty(name='Possible Cyclic dependency', description='Curent Hair setup may form cyclic dependency:\n - Guide curve depends on your Base mesh,\n - Base mesh (with Hair System attached) depends on Guide Curve\nInterpolation using current Guide Curve may not work. To fix use: Ctrl+Shift+H ->  Transfer Hair System -> To Guide (or to Empty Mesh)', default=False)
    hair_base_mesh: bpy.props.StringProperty(name='Hair Base Mesh', description='Restore original Hair System mesh', default='')
    is_convert_result: bpy.props.BoolProperty(name='Is convert result', description='This object is result of converting Hair System to mesh', default=False)



class JBone(bpy.types.PropertyGroup):
    def up_color(self, context):
        arma = self.id_data
        obj = context.active_object
        bone_colls = self.id_data.path_resolve('collections')
        # find collection that holds this physics props, and update bone colors in it
        for coll in bone_colls:
            if coll.physics_props == self:
                for bone in coll.bones:
                    pbone = obj.pose.bones.get(bone.name)
                    set_bone_col(pbone, self.bone_color)

    name: bpy.props.StringProperty(name='Name', description='Jiggle physics settings name', default='Jiggle')
    mass: bpy.props.FloatProperty(name='Mass', description='Mass of each spring segment', default=0.2, min=0.01, soft_max=10)
    stiffness: bpy.props.FloatProperty(name='Stiffness', description='Stiffnes of each spring segment', default=10., min=0.01, soft_max=100)
    gravity: bpy.props.FloatProperty(name='Gravity', description='Gravity strength', default=9.81, min=0.0, max=20)
    damping: bpy.props.FloatProperty(name='Damping', description='Damping of each spring segment', default=0.2, min=0.01, soft_max=1)
    bone_color: bpy.props.FloatVectorProperty(name='Color', description='Color of each spring segment', default=(1.0, 1.0, 1.0), min=0.0, max=1.0, subtype='COLOR', size=3, update=up_color)
    is_jiggle: bpy.props.BoolProperty(name='Is jiggle', description='Is Bone Collection holding jiggle data?', default=False)


class JiggleBoneProps(bpy.types.PropertyGroup):
    physics_props: bpy.props.CollectionProperty(type=JBone)
    # time_step: bpy.props.FloatProperty(name='Time Step', description='Lower time step gives smoother animation', default=0.04, min=0.01, max=1 )
    active_index: bpy.props.IntProperty(name="Active index", description='', default=1, min=0, max=100 )


UNIQUE_BOX_ITEMS  = []
def unique_uv_box_items(self, context):
    obj = context.active_object
    first_mat = obj.material_slots[0].material
    uvBoxes = len(first_mat.ht_props.hair_uv_points)
    global UNIQUE_BOX_ITEMS
    if len(UNIQUE_BOX_ITEMS) != uvBoxes:
        UNIQUE_BOX_ITEMS = [(str(i), f'UV box {i}','') for i in range(uvBoxes)]
    return UNIQUE_BOX_ITEMS

class MeshShortHairProps(bpy.types.PropertyGroup):
    def assign_uv_regions(self, context):
        # DONE: run vgroup to uv update
        # if self.vg_name and  len(self.uv_regions) > 0:
        update_all_uv_new(context.active_object)

    vg_name: bpy.props.StringProperty(name='Vertex Group', description='Vertex group that will mask UV regions',update=assign_uv_regions)
    mute: bpy.props.BoolProperty(name='Mute', description='If checked, disable current Vertex Group UVs', default=False, update=assign_uv_regions)
    uv_regions: bpy.props.EnumProperty(name='UV Region ID', description='UV regions that will be assigned vertex group', items=unique_uv_box_items,
                                       options={'ENUM_FLAG'},
                                       update=assign_uv_regions)

class MeshShortHairUVProps(bpy.types.PropertyGroup):
    short_hair_vg_uvs: bpy.props.CollectionProperty(type=MeshShortHairProps)
    active_index: bpy.props.IntProperty(name="Active index", description='', default=1, min=0, max=100 )
    seed: bpy.props.IntProperty(name='Seed', description='Random seed', default=0, min=0, max=1000)

class MeshHTPropertyGroup(bpy.types.PropertyGroup):
    short_hair: bpy.props.PointerProperty(type=MeshShortHairUVProps)

class ObjHTPropertyGroup(bpy.types.PropertyGroup):
    def assign_uv_regions(self, context):
        obj = self.id_data
        if obj.type == 'CURVES' and self.use_auto_uv:
            auto_update_all_curves_uv_regions(obj)

    is_curve: bpy.props.IntProperty(name='Is curve',default=0, min=0, max=1) # for OffsetNode attribute
    short_hair_props: bpy.props.PointerProperty(type=ObjShortHairProps)
    shell_hair_props: bpy.props.PointerProperty(type=ObjShellHairProps)
    hair_settings: bpy.props.PointerProperty(type=hair_ribbons_props)
    curly_hair_settings: bpy.props.PointerProperty(type=curly_hair_settings)
    rand_tilt_settings: bpy.props.PointerProperty(type=rand_tilt_settings)
    radius_from_length_settings: bpy.props.PointerProperty(type=radius_from_length_settings)
    target_obj: StringProperty(name="Target", description="Extra target object for operation (overrides Curve Surface input)")

    bone_gen_props: bpy.props.PointerProperty(type=ObjBoneGenProps)

    def filter_armature(self, obj):
        return obj.type == 'ARMATURE'

    target_rig: bpy.props.PointerProperty(name='Target rig', description="If set, generated bones will be automatically assigned to that armature object", type=bpy.types.Object, poll=filter_armature)
    pin_bone: bpy.props.StringProperty(name='Parent bone', description='Generated bones will be automatically set as child of this bone (in most cases it is same as Head bone)', default='')

    used_uv_pick: bpy.props.BoolProperty(name='had uv sample', description='', default=False)
    used_weight_pick: bpy.props.BoolProperty(name='had weight sample', description='', default=False)

    use_auto_uv: bpy.props.BoolProperty(name='Auto UV Region Distribute', description='Automaticaly assign UV regions to hair strands when generating/updating hair.\nCustomize UV regions assignments rules in: UV Editor > Right Sidebar > Hair Tool UV', default=True, update=assign_uv_regions)
    uv_seed: IntProperty(name="Seed", description='Randomize UV assigment', default=0, min=0, max=1000, update=assign_uv_regions)

    jiggle_props: bpy.props.PointerProperty(type=JiggleBoneProps)

    # Hair System Things
    hair_nodes: bpy.props.PointerProperty(type=HairNodesProps)

    active_selection_id: bpy.props.IntProperty(name="Active Group ID", description='Currently active Group Id - can be assigned to selected strands', default=0, min=0, soft_max=100 )
    hidden_strands_obj: bpy.props.PointerProperty(type=bpy.types.Object)

    stored_matrix: bpy.props.FloatVectorProperty(name ='Backup position',options={'HIDDEN'}, size=16) # for 'UV Flatten' operator

class JiggleScnProps(bpy.types.PropertyGroup):
    sim_is_running: bpy.props.BoolProperty(name='Jiggle sim on', description='', default=False)
    # jiggle_frame_step: bpy.props.BoolProperty(name='Jiggle frame step', description='', default=False)
    sim_mode: bpy.props.EnumProperty(name='Mode', description='Sim Mod: Frame Change - can be baked into animation, Realtime - can be used for realtime simulation',
        items=[
            ('FRAME_CHANGE', 'Animated', 'Good for jiggle baking and rendering animation. Only works when animation is playing'),
            ('REALTIME', 'Interactive', 'Good for in viewport interaction with jiggle bones')
        ], default='REALTIME')

FOLDER_ITEMS = {}  # for keeping reference in python to items - fix enum bug in blender
def folder_items(self, context):
    global FOLDER_ITEMS
    pref = get_addon_preferences()
    lib_path_len = len(pref.lib_paths[pref.lib_active_index].directory)
    curr_sub_dir = context.scene.ht_props.lib_relative_sub_path
    lib_extended_path = os.path.join(pref.lib_paths[pref.lib_active_index].directory, curr_sub_dir)
    if not hair_lib.FORCE_UPDATE_FOLDERS and lib_extended_path in FOLDER_ITEMS.keys():
        return FOLDER_ITEMS[lib_extended_path]

    # print(f'Updating. Force update was {FORCE_UPDATE_FOLDERS}')
    items = []
    if os.path.exists(lib_extended_path):
        for fn in os.listdir(lib_extended_path):
                file_path = os.path.join(lib_extended_path, fn)
                if os.path.isdir(file_path):
                    relative_sub_dir = file_path[lib_path_len:]
                    items.append((relative_sub_dir, '+', fn))
        FOLDER_ITEMS[lib_extended_path] = items
        hair_lib.FORCE_UPDATE_FOLDERS = False
        return FOLDER_ITEMS[lib_extended_path]
    else:
        return items




ATTR_CACHED = {}
class ScnHTPropertyGroup(bpy.types.PropertyGroup):
    def filter_mesh(self, obj):
        return obj.type == 'MESH'

    def reset_offset(self, context):
        self.select_offset = 0

    def draw_strands_up(self, context):
        if self.strands_overlay:
            if on_draw_guide_overlay_depsgraph_update not in bpy.app.handlers.depsgraph_update_post:
                bpy.app.handlers.depsgraph_update_post.append(on_draw_guide_overlay_depsgraph_update)
        else:
            if on_draw_guide_overlay_depsgraph_update in bpy.app.handlers.depsgraph_update_post:
                bpy.app.handlers.depsgraph_update_post.remove(on_draw_guide_overlay_depsgraph_update)
        draw_new_hair(context)

    def attr_set(self, active_obj, value_name, attr_name, default_value, mode = 'ADD'):
        if not active_obj or active_obj.type != 'CURVES':
            return
        # sel_pts_bool = get_edit_selected_points_bool(active_obj)
        sel_mask_float = get_any_type_fselection(bpy.context, active_obj)
        global ATTR_CACHED
        if active_obj.name not in ATTR_CACHED.keys() or ATTR_CACHED['attr_name']!= attr_name or ATTR_CACHED['sel'] != sel_mask_float:
            print('up cache')
            ATTR_CACHED.clear()
            ATTR_CACHED[active_obj.name] = get_create_attrib_cache(active_obj, attr_name, default_value=default_value)
            ATTR_CACHED['sel'] = sel_mask_float
            ATTR_CACHED['attr_name'] = attr_name

        affected_attr = active_obj.data.attributes.get(attr_name)
        if not affected_attr:
            affected_attr = active_obj.data.attributes.new(name=attr_name, type='FLOAT', domain='POINT')
        # sel_count = np.sum(sel_pts_bool)
        # avg_rad = sum([affected_attr.data[pt_idx].value for pt_idx,sel in enumerate(sel_pts_bool) if sel]) / sel_count
        # sel_pts_mask_f = get_edit_selected_points_float(active_obj) # selection can be float now.
        sel_pts_mask_f = sel_mask_float
        current_val = getattr(self, value_name)
        if mode == 'SET':
            for pt_idx, attr_data in enumerate(affected_attr.data):
                old_val = ATTR_CACHED[active_obj.name][pt_idx]
                attr_data.value = lerp(old_val, current_val, sel_pts_mask_f[pt_idx])
        elif mode == 'MUL':
            for pt_idx, attr_data in enumerate(affected_attr.data):
                old_val = ATTR_CACHED[active_obj.name][pt_idx]
                attr_data.value = lerp(old_val, old_val * current_val, sel_pts_mask_f[pt_idx])
        elif mode == 'ADD':
            for pt_idx, attr_data in enumerate(affected_attr.data):
                old_val = ATTR_CACHED[active_obj.name][pt_idx]
                attr_data.value = lerp(old_val, old_val + current_val, sel_pts_mask_f[pt_idx])

        self[value_name] = default_value # after done dragging slider- reset to default

    def up_rad_mult(self, context):
        self.attr_set(context.active_object, 'radius_mult', 'radius', default_value=1, mode='MUL')

    def up_tilt_add(self, context):
        self.attr_set(context.active_object, 'tilt_add', 'tilt', default_value=0, mode='ADD')

    def up_set_rad(self, context):
        self.attr_set(context.active_object, 'radius_set', 'radius', default_value=1, mode='SET')

    def up_set_tilt(self, context):
        self.attr_set(context.active_object, 'tilt_set', 'tilt', default_value=1, mode='SET')

    def up_system_color(self, context):
        bpy.ops.curves.set_color_attribute('INVOKE_DEFAULT', attr_name='SystemColor', color=self.system_color, default_domain='CURVE', default_value=[0,0,0,1])


    lib_relative_sub_path: bpy.props.StringProperty(name="Folder Path", subtype='DIR_PATH', description='Holds library subfolder', default="")
    lib_previews: bpy.props.EnumProperty(items=enum_previews_from_directory_items, description='Holds previews')
    export_merged_name: bpy.props.StringProperty(name="Merged name", description="When saving multiple hairs, define output name", default="Merged")
    save_with_screen_grab: bpy.props.BoolProperty(name="Save with Screen Grab", description="Save with Screen Grab", default=True)
    unfold_preset_creation: bpy.props.BoolProperty(name="Create Preset", description="Create Preset", default=False)
    bool_folders: bpy.props.EnumProperty(name='Folders select', description='Checkbox for folders to perform operation on',
                                         items=folder_items,
                                         options={'ENUM_FLAG'})

    hair_bake_settings: bpy.props.PointerProperty(type=HairBakeSettings)
    hair_drawing: bpy.props.PointerProperty(type=ModalDrawingSettings)
    active_end: bpy.props.EnumProperty(name='Active end',
                                      items=(
                                          ('ROOTS', 'Model Roots', ''),
                                          ('TIPS', 'Model Tips', '')),
                                       default='TIPS', update=reset_offset)

    use_proportional_edit:bpy.props.BoolProperty(name="Use Proportional Edit", description="Mask out effect applied to hair strands along their length", default=True)
    proportional_edit_falloff: bpy.props.EnumProperty(name='Proportional Edit Falloff',
        items=(
            ('SMOOTH'        , 'Smooth', '', 'SMOOTHCURVE',1),
            ('SPHERE'        , 'Sphere', '', 'SPHERECURVE',2),
            ('ROOT'          , 'Root', '', 'ROOTCURVE',3),
            ('INVERSE_SQUARE', 'Inverse Square', '', 'INVERSESQUARECURVE',4),
            ('SHARP'         , 'Sharp', '', 'SHARPCURVE',5),
            ('LINEAR'        , 'Linear', '', 'LINCURVE',6),
            ('CONSTANT'      , 'Constant', '', 'NOCURVE',7),
        ),
        default='SMOOTH')



    higlight_bias: bpy.props.FloatProperty(name="Draw bias", description="Draw strand points closer to camera", min=0.0, soft_max=1.0, default=0.5, subtype='FACTOR')
    higlight_range: bpy.props.IntProperty(name='Higlight adjacent range', description='', default=1, min=0, max=2)
    select_offset: bpy.props.IntProperty(name='Selection Offset', description='', default=0, min=0, max=32)
    clump_falloff: bpy.props.FloatProperty(name="Strand stiffness", description="Strand will resist deformation more", default=0.5, min=0, max=1)
    influence_rad: bpy.props.FloatProperty(name="Proportianl Size", description="Proportianl Size", default=0.5, min=0.001, soft_max=10)
    circle_sel_rad2d: bpy.props.IntProperty(name="Circle Select Size", description="Proportianl Size", default=100, min=10, soft_max=500)
    align_tilt: bpy.props.BoolProperty(name='Align Tilt', description='', default=False)
    move_mode: bpy.props.EnumProperty(name='Mode',
                                       items=(
                                           ('FK', 'Normal (FK)', 'Transform hair with proportional fallof'),
                                           ('ROOT', 'Root', 'Transform hair around theirs root points'),
                                           ('IK', 'Chain (IK)', 'Move hair as chains')),
                                       default='IK')
    use_collisions: bpy.props.BoolProperty(name='Use Collisions', description='', default=True)
    collison_offset: bpy.props.FloatProperty(name="Collision Offset", description="Collision Offset distance", default=0.01, min=0.0, soft_max=1.0, precision=3,subtype='DISTANCE')
    use_custom_collider: bpy.props.BoolProperty(name='Use Custom Collider', description='', default=False)
    custom_collider: bpy.props.PointerProperty(name='Custom Collider', description='Custom collider object', type=bpy.types.Object,poll=filter_mesh)


    # Jiggle props
    jiggle_props: bpy.props.PointerProperty(type=JiggleScnProps)

    # For Hair Systems
    strands_overlay: BoolProperty(name="Parent strands overlay", description="Draw orange overlay on top of parent strands", default=False, update=draw_strands_up)

    radius_mult: bpy.props.FloatProperty(name="Radius Multiplier", description="Multipley selected pointns radius by this this value", default=1.0, min=0.001, soft_max=10,  update=up_rad_mult) #, get=get_rad ,set=set_rad, update=up_rad, get=get_rad
    tilt_add: bpy.props.FloatProperty(name="Add Tilt", description="Active Tilt", default=0.0, soft_min=-3.14, soft_max=3.14, update=up_tilt_add)

    radius_set: bpy.props.FloatProperty(name="Set Radius", description="Multipley selected pointns radius by this this value", default=1.0, min=0.001, soft_max=10,  update=up_set_rad) #, get=get_rad ,set=set_rad, update=up_rad, get=get_rad
    tilt_set: bpy.props.FloatProperty(name="Set Tilt", description="Active Tilt", default=0.0, soft_min=-3.14, soft_max=3.14, update=up_set_tilt)

    # for hair system popup panel
    show_profile: bpy.props.BoolProperty(name="Profile Section", description="Show Profile", default=False)

    # system color setter
    system_color: bpy.props.FloatVectorProperty(name="System Color", description="Set color to 'SystemColor' attribute for selected strands", default=(0.0, 0.0, 0.0, 1.0), subtype='COLOR', min=0.0, max=1.0, size=4, update=up_system_color)

    # color presets for uv boxes
    ht_uv_palette: bpy.props.PointerProperty(type=bpy.types.Palette, description="Use 'Hair UV' operator to assign colors to UV regions (with [S] key)")


class WMHTPropertyGroup(bpy.types.PropertyGroup):
    pass


class MatHTPropertyGroup(bpy.types.PropertyGroup):
    hair_uv_points: bpy.props.CollectionProperty(type=hairUVPoints)
    linked_mesh_mat: bpy.props.PointerProperty(type=bpy.types.Material)



# is this used?
class CurveHTPropertyGroup(bpy.types.PropertyGroup):
    uv_seed: IntProperty(name="UV Seed", default=20, min=1, max=1000)

class PBoneHTPropertyGroup(bpy.types.PropertyGroup):
    jiggle_index: bpy.props.IntProperty(name="Jiggle Index", default=-1, min=-1, max=32)
    # jiggle_collision  - used in both b3.6 and b4.0
    jiggle_collision: bpy.props.BoolProperty(name='Add Collision', description='Other bones will collide with this bone.\nCollision shape is described by bone envelope - Tip and Head Radius.\nCollision margin is described by bone envelope size', default=False)
    friction: bpy.props.FloatProperty(name='Friction', description='', default=0.4, min=0, max=10)


def register_properties():
    bpy.types.Object.ht_props = bpy.props.PointerProperty(type=ObjHTPropertyGroup)
    bpy.types.Mesh.ht_props = bpy.props.PointerProperty(type=MeshHTPropertyGroup)
    bpy.types.Scene.ht_props = bpy.props.PointerProperty(type=ScnHTPropertyGroup)
    bpy.types.Material.ht_props = bpy.props.PointerProperty(type=MatHTPropertyGroup)
    bpy.types.Curve.ht_props = bpy.props.PointerProperty(type=CurveHTPropertyGroup)
    bpy.types.PoseBone.ht_props = bpy.props.PointerProperty(type=PBoneHTPropertyGroup)
    bpy.types.WindowManager.ht_props = bpy.props.PointerProperty(type=WMHTPropertyGroup)
    if bpy.app.version >= (4, 0, 0):
        bpy.types.BoneCollection.physics_props = bpy.props.PointerProperty(type=JBone)



def unregister_properties():
    if bpy.app.version >= (4, 0, 0):
        del bpy.types.BoneCollection.physics_props
    del bpy.types.Object.ht_props
    del bpy.types.Mesh.ht_props
    del bpy.types.Scene.ht_props
    del bpy.types.Material.ht_props
    del bpy.types.Curve.ht_props
    del bpy.types.PoseBone.ht_props
    del bpy.types.WindowManager.ht_props
