# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
from .utils.general_utils import addon_name_lowercase,get_addon_preferences
import rna_keymap_ui

from .drawing_curves import HTOOL_PT_DrawingCurvesPanel
from .hair_baking.hair_bake_ui import HTOOL_PT_Hair_Panel_Bake
from .shell_hair import get_shell_hair_mod
from .hair_baking.hair_geometry_nodes_drawing import update_hairs_pos
from .hair_baking.hair_geometry_nodes_shared import get_profile_mod, get_profile_type_node


# from .hair_baking.hair_geometry_nodes import HTOOL_PT_HairSystemPanel
from pathlib import Path
import os

HAIR_IC = 'NONE' if bpy.app.version >= (3, 2, 0) else 'HAIR'


# define submenu for all the radius affecting
class HTOOL_MT_CurvesContextMenu(bpy.types.Menu):
    bl_idname = "HTOOL_MT_CurvesContextMenu"
    bl_label = "Curves Context Menu"

    def draw(self, context):
        layout = self.layout

        layout.operator("curves.subdivide_curve_segment")
        layout.operator("curves.extrude_curve_ending")
        layout.operator("curves.hair_duplicate")
        layout.operator("curves.separate_selected")
        if bpy.app.version < (4, 2, 0): # use HTool version
            layout.operator("curves_ht.switch_direction")
        else: # in 4.2 blender has its own version
            layout.operator("curves.switch_direction")

        layout.separator()
        layout.operator_context = "INVOKE_DEFAULT"

        layout.operator("curves.editmode_adjust_tilt_rad", text='Adjust Radius').mode = 'RADIUS'
        op = layout.operator("curves.set_attribute", text='Set Radius')
        op.attr_name = 'radius'
        op.default_value = 1
        op.value = 1

        layout.separator()
        layout.operator("curves.editmode_adjust_tilt_rad", text='Adjust Tilt').mode = 'TILT'
        op = layout.operator("curves.set_attribute", text='Set Tilt')
        op.attr_name = 'tilt'
        op.default_value = 0
        op.value = 0


# define submenu for all the radius affecting
class HTOOL_MT_SmoothAttributeSubmenu(bpy.types.Menu):
    bl_idname = "HTOOL_MT_SmoothAttributeSubmenu"
    bl_label = "Smooth Attribute"

    def draw(self, context):
        layout = self.layout
        layout.operator("curves.curves_smooth_attr", text='Smooth Radius').attr_name = 'radius'  # , icon='MOD_SMOOTH'
        layout.operator("curves.curves_smooth_attr", text='Smooth Tilt').attr_name = 'tilt'  # , icon="MOD_SMOOTH"
        layout.operator("curves.curves_smooth_attr", text='Smooth Position').attr_name = 'position'  # , icon="MOD_SMOOTH"
        # layout.operator("curve.radius_by_length", icon='MOD_LENGTH')


def rad_curve_box(context, obj, box, draw_label=True):
    col = box.column(align=True)
    if draw_label:
        col.label(text="Radius")
    # if context.mode in {'EDIT_CURVES', 'SCULPT_CURVES'}:
    col.prop(context.scene.ht_props, 'radius_mult')
    col.prop(context.scene.ht_props, 'radius_set')
    # op = col.operator("curves.set_attribute", text='Set Radius')
    # op.attr_name = 'radius'
    # op.default_value = 1
    # op.value = 1

    col.operator("curves.curve_taper")  # , icon='SHARPCURVE'
    # col.operator("curve.radius_by_length", icon='MOD_LENGTH')


def tilt_curve_box(context, obj, box, draw_label=True):
    col = box.column(align=True)
    if draw_label:
        col.label(text="Tilt")
    # if context.mode in {'EDIT_CURVES', 'SCULPT_CURVES'}:
    col.prop(context.scene.ht_props, 'tilt_add')
    col.prop(context.scene.ht_props, 'tilt_set')
    # op = col.operator("curves.set_attribute", text='Set Tilt')
    # op.attr_name = 'tilt'
    # op.default_value = 0
    # op.value = 0
    if obj.mode != 'EDIT':
        col.operator("object.curves_align_tilt")  # , icon="SNAP_NORMAL"
        col.prop_search(obj.ht_props, "target_obj", context.scene, "objects")  # ,icon="SNAP_NORMAL"
    col.operator("curves.curves_randomize_tilt")  # , icon="FORCE_MAGNETIC"


def utils_curve_box(context, obj, box, draw_label=True):
    col = box.column(align=True)
    if draw_label:
        col.label(text="Utils")
    # col.operator("object.embed_roots") # , icon="MOD_SCREW"
    # col.prop_search(obj.ht_props, "target_obj", context.scene, "objects") # ,icon="SNAP_NORMAL"

    col.menu("HTOOL_MT_SmoothAttributeSubmenu", icon="THREE_DOTS")
    col.operator("curves.hair_straighten")  # , icon="STRANDS"
    col.operator("curves.adjust_length")  # , icon='MOD_LENGTH'
    col.operator("curves.curve_resample")  # , icon="PARTICLE_POINT"
    col.operator("curves.curve_simplify")  # , icon="PARTICLE_POINT"


def other_curve_box(context, obj, box, draw_label=True):
    col = box.column(align=True)  # 'new' way - using geo nodes to do stuff
    if draw_label:
        col.label(text="Other")
    # col.operator("hair_tool.resample_curves", icon="PARTICLE_POINT") # using geo nodes mod
    if context.mode == 'OBJECT':
        col.operator('curves.join_strands')  # using geonodes # , icon='FULLSCREEN_EXIT'
    col.operator('curves.separate_selected')  # using geonodes # , icon='FULLSCREEN_ENTER'
    # col.operator('hair_tool.remove_strands', icon='TRASH') # using geonodes
    col.operator("curves.slice_with_mesh")  # , icon="RESTRICT_INSTANCED_ON"

    row = col.row(align=True)
    row.prop(context.scene.ht_props, 'system_color')

    col = box.column(align=True)
    sub_r = col.row(align=True)
    sub_r.operator('curves.hide_selection')
    sub_r.operator('curves.unhide_selection')

    col = box.column(align=True)
    col.operator("hsystem.add_ao_mod")  # , icon="VPAINT_HLT"


def profile_ops(context, box, draw_label=True):
    obj = context.active_object
    profile_mod = get_profile_mod(obj)
    col = box.column(align=True)
    if draw_label:
        col.label(text='Profile')
    split = col.split(factor=0.8, align=True)
    profile_txt = 'Edit Profile' if profile_mod else 'Add Profile'
    split.operator("object.generate_profile", text=profile_txt)  # , icon="OUTLINER_OB_SURFACE"
    if not profile_mod:
        col.operator("hair_profile.paste_settings", text='Paste Profile')  # , icon='PASTEDOWN'
        return

    split.operator('hair_system.make_profile_unique', text=f'{profile_mod.node_group.users}', emboss=True)
    profile_node = get_profile_type_node(profile_mod.node_group)
    if profile_node:
        input_keys = profile_node.inputs.keys()
        if 'Global Scale' in input_keys:
            col.prop(profile_node.inputs['Global Scale'], 'default_value', text='Global Scale')
        elif 'Width' in input_keys:
            col.prop(profile_node.inputs['Width'], 'default_value', text='Width')
        elif 'Profile Width' in input_keys:
            col.prop(profile_node.inputs['Profile Width'], 'default_value', text='Profile Width')

        if 'Resolution' in input_keys:
            col.prop(profile_node.inputs['Resolution'], 'default_value', text='Resolution')
    row = col.row(align=True)
    row.operator("hair_profile.copy_settings", text='Copy')  # , icon='COPYDOWN'
    row.operator("hair_profile.paste_settings", text='Paste')  # , icon='PASTEDOWN'
    # col.operator("hair_profile.pick_last_settings")


def attributes_ops(context, box, draw_label=True):
    col = box.column(align=True)
    if draw_label:
        col.label(text='Attributes: ')
    is_editing = context.scene.ht_props.get('editing_attribute')
    if is_editing:
        col.operator_menu_enum("curves.attribute_to_selection", 'selected_attribute', icon='CHECKMARK', text='Editing: ' + is_editing)  # , icon="EVENT_A"
    else:
        col.operator_menu_enum("curves.attribute_to_selection", 'selected_attribute')  # , icon="EVENT_A"
    row = col.row(align=True)
    row.operator("curves.selection_to_attribute")
    row.operator("curves.cancel_attribute_edit", icon='PANEL_CLOSE', text='')
    if is_editing:
        col.operator("curves.curves_smooth_attr", text='Smooth: ' + is_editing).attr_name = '.selection'  # , icon='MOD_SMOOTH'
    else:
        col.operator_menu_enum("curves.curves_smooth_attr", "menu_selected_attribute").attr_name = ''  # , icon='MOD_SMOOTH'
    op = col.operator("curves.curve_map_attribute", text="Selection Mask (Curve)")  # , icon='MOD_SMOOTH'
    op.attr_name = '.selection'
    op.force_attr_type = True
    # col.operator("curves.set_curve_resolution") # , icon="PARTICLE_POINT"
    # op = col.operator("curves.set_color_attribute", text='Set Color') # XXX:  point domain on curve > to ribbon - will give wrong result (would have to sample curve to profile..?)
    # op.attr_name = 'HairColor'
    # op.default_domain = "POINT"
    # op.default_value = [0,0,0,1]


def generate_ops(context, obj, box, draw_label=True):
    col = box.column(align=True)
    if draw_label:
        col.label(text='Generate')

    if context.mode == 'OBJECT':
        if obj.ht_props.hair_nodes.is_using_gnodes:
            col.operator("screen.draw_hsys_popup")
        else:
            op = col.operator_menu_enum("object.add_hair_system", "mode")
            op.draw_popup = True

    if obj.type == 'CURVES':
        # col.operator("object.curve_curl", icon="LINKED", text=curls_txt)
        # if obj.ht_props.target_obj:
        # col.operator("object.curves_from_grid", icon="MESH_GRID")
        # col.operator("add_hair_system").mode = 'GRID_GUIDE'
        draw_props = context.scene.ht_props.hair_drawing
        row = col.row(align=True)
        row.prop(draw_props, 'draw_strands', emboss=True, toggle=True)
        row.operator('object.clear_cache', text='', icon="FILE_REFRESH")

    elif obj.type == 'MESH':  # for non ribbon kind of  mesh
        if context.mode == 'EDIT_MESH':
            if obj.ht_props.short_hair_props.is_short_hair:
                col.operator("hair.set_short_hair_uv_region")  # , icon="MATERIAL"
            row = col.row(align=True)
            row.operator("object.ribbon_to_bones")  # , icon="BONE_DATA"
            row.operator("object.precalc_islands_cache", icon="FILE_REFRESH", text='')
            col.prop(obj.ht_props, 'target_rig')
            if obj.ht_props.target_rig and obj.ht_props.target_rig.type == 'ARMATURE':
                col.prop_search(obj.ht_props, 'pin_bone', obj.ht_props.target_rig.data, 'bones')
        else:
            if obj.ht_props.hair_settings.hair_mesh_parent:  # for ribbon mesh
                col.operator("object.ribbon_convert", icon="FILE_REFRESH")

            if obj.ht_props.short_hair_props.hair_base_mesh:
                pass
            if obj.ht_props.shell_hair_props.hair_base_mesh:
                pass
            if obj.ht_props.hair_nodes.hair_base_mesh:
                pass
            if obj.ht_props.short_hair_props.is_short_hair:
                pass
            if get_shell_hair_mod(obj):
                pass
            else:
                # col.operator("object.curves_from_grid", icon="MESH_GRID")
                if obj and obj.type == 'MESH' and not obj.ht_props.hair_nodes.is_using_gnodes:
                    col.operator("object.curves_from_grid_new")  # , icon="MESH_GRID"
                    col.operator("curves.hairsys_with_guide_mesh")  # , icon="MESH_GRID"

                row = col.row(align=True)
                row.prop(context.scene.ht_props.hair_drawing, 'draw_strands', emboss=True, toggle=True)  # , icon='PARTICLEMODE'
                row.operator('object.clear_cache', text='', icon="FILE_REFRESH")

        if context.mode == 'OBJECT':
            row = col.row(align=True)
            row.operator("object.setup_short_hair")  # , icon=HAIR_IC
            row.operator("hair.update_short_hair_node_group", icon="FILE_REFRESH", text='')
            row = col.row(align=True)
            row.operator("object.setup_shell_hair")  # , icon="MOD_WAVE"
            row.operator("hair.update_shell_hair_node_group", icon="FILE_REFRESH", text='')


def convert_ops(context, obj, box, draw_label=True):
    col = box.column(align=True)
    if draw_label:
        col.label(text='Convert')
    if get_profile_mod(obj):
        col.operator("hair_system.convert_to_mesh")  # , icon="FILE_REFRESH"

    if obj.ht_props.hair_nodes.is_using_gnodes:  # Hair system - for mesh or curve
        col.operator("hair_system.finalize")  # , icon="SOLO_ON"
        col.operator("hair_system.transfer_to_object")  # , icon="MOD_BOOLEAN"
        col.operator('hair_system.bake_whole_system')  # , icon='CHECKMARK'
        return
        # col.operator("curves.curve_to_mesh_ribbon") # , icon="FILE_REFRESH" # replaced by hair_system.convert_to_mesh
    elif obj.type == 'MESH':  # for non ribbon kind of  mesh
        if context.mode == 'OBJECT':
            col.operator("object.create_edge_strips")

            if obj.ht_props.hair_settings.hair_mesh_parent:  # for ribbon mesh
                col.operator("object.ribbon_convert", icon="FILE_REFRESH")

            if obj.ht_props.short_hair_props.hair_base_mesh:
                col.operator("object.restore_short_hair_base_mesh")  # , icon="FILE_REFRESH"
            if obj.ht_props.shell_hair_props.hair_base_mesh:
                col.operator("object.restore_shell_hair_base_mesh")  # , icon="FILE_REFRESH"
            if obj.ht_props.hair_nodes.hair_base_mesh:
                col.operator("hair_system.restore_base_mesh")  # , icon="FILE_REFRESH"
            if obj.ht_props.short_hair_props.is_short_hair:
                col.operator("hair.convert_short_hair_to_mesh")  # , icon="FILE_REFRESH"
            if get_shell_hair_mod(obj):
                col.operator("hair.convert_shell_hair_to_mesh")  # , icon="FILE_REFRESH"
            else:
                pass

            # if not obj.ht_props.hair_settings.hair_mesh_parent:
            if len(context.selected_objects) > 1 and any([o.type == 'CURVE' for o in context.selected_objects]):
                col.operator("object.hair_from_curves")  # , icon="FILE_REFRESH"
            for mod in context.object.modifiers:
                if mod.type == 'PARTICLE_SYSTEM' and mod.show_viewport and mod.particle_system.settings.type == 'HAIR':  # use first visible
                    col.operator("object.hair_to_curves")  # , icon=HAIR_IC
                    break

    if context.mode == 'OBJECT':
        # if len(context.selected_objects) > 1:
        #     col.operator("object.hair_from_curves") # , icon="FORCE_CURVE"
        col.operator("object.finalize_hair")  # , icon="SOLO_ON"


def material_uv_ops(obj, box, draw_label=True):
    col = box.column(align=True)
    # profile_mod = get_profile_mod(obj)
    if draw_label:
        col.label(text='Hair Profile Material')
    col.operator('hair.load_default_mat')  # , icon='IMPORT'
    # col.operator("object.curve_uv_refresh") # , icon="MATERIAL"
    col.prop(obj.ht_props, 'use_auto_uv', toggle=True)  # , icon='MOD_UVPROJECT'

    # if profile_mod:
    #     row = col.row(align=True)
    #     row.prop_search(profile_mod, f'["{prof_mat_input_name}"]', bpy.data, 'materials', text='', icon='MATERIAL_DATA')
    #     mat = profile_mod[prof_mat_input_name]
    #     if mat and mat.users > 1:
    #         row.alignment = 'RIGHT'
    #         row.operator('hair_system.make_material_unique', text=f'[{mat.users-1}]', emboss=True)

    # col = col.column(align=True)
    # col.label(text='Hair profile UVs')
    # if profile_mod:
    #     row = col.row(align=True)
    #     row.prop_search(profile_mod, f'["{uv_owner_input_name}"]', bpy.data, 'objects', text='UV Data', icon='OBJECT_DATA')
    #     uv_owner = profile_mod[uv_owner_input_name]
    #     if uv_owner and uv_owner.users > 2:
    #         row.alignment = 'RIGHT'
    #         row.operator('hair_system.make_uv_unique', text=f'[{uv_owner.users-1}]', emboss=True)

    op = col.operator("hair_system.set_uv_regions_new")  # , icon="MATERIAL"
    op.force_use_reg_deformer = False


def mesh_ops(context, obj, box, draw_label=True):
    if draw_label:
        box.label(text='Mesh operations')
    col = box.column(align=True)
    if context.mode == 'EDIT_MESH':
        col.separator()
    else:
        col.operator("object.ribbon_weight")  # , icon="WPAINT_HLT"
        col.operator("hsystem.add_ao_mod")  # , icon="VPAINT_HLT"
        col.operator("object.ribbon_vert_color_grad")  # , icon="VPAINT_HLT"
        col.operator("object.uvgradient_bake")  # , icon="NODE_TEXTURE"
        col.operator("object.ribbon_vert_color_random")  # , icon="VPAINT_HLT"
        col.prop_search(obj.ht_props, "target_obj", context.scene, "objects", icon="SNAP_NORMAL")
        split = col.split(factor=0.05)
        split.enabled = obj.ht_props.target_obj is not None
        # if obj.ht_props.target_obj:
        split.label(text='')
        scol = split.column(align=True)
        scol.operator("object.uv_sample_from_target")  # , icon="VPAINT_HLT"
        scol.operator("object.color_sample_from_target")  # , icon="COLOR"
        scol.operator("object.weight_sample_from_target")  # , icon="MOD_VERTEX_WEIGHT"


class HT_RightSiderPanel:
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Hair Tool'
    bl_options = {'DEFAULT_CLOSED'}


class HTOOL_PT_VIEW3D_PT_Hair_Panel(HT_RightSiderPanel, bpy.types.Panel):
    bl_label = "Hair Operators"
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel"

    # bl_context = 'objectmode' #
    # bl_options = {'HIDE_HEADER'}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type in {'MESH', 'CURVES'} and context.mode in {'OBJECT',  'EDIT_CURVES', 'SCULPT_CURVES'}

    def draw(self, context):
        layout = self.layout
        # layout.operator("object.cleanup_hair", icon="CANCEL")


class HTOOL_PT_VIEW3D_PT_Hair_Panel_Convert(HT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel_Convert"
    bl_label = "Convert"
    bl_parent_id = "HTOOL_PT_VIEW3D_PT_Hair_Panel"
    bl_order = 5

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj  # and (obj.type == 'MESH' and  get_profile_mod(obj)) or obj.type == 'CURVES'

    def draw(self, context):
        layout = self.layout
        convert_ops(context, context.active_object, layout, draw_label=False)


class HTOOL_PT_VIEW3D_PT_Hair_Panel_Generate(HT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel_Generate"
    bl_label = "Generate"
    bl_parent_id = "HTOOL_PT_VIEW3D_PT_Hair_Panel"
    bl_order = 10

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj  # and (obj.type == 'MESH' and  get_profile_mod(obj)) or obj.type == 'CURVES'

    def draw(self, context):
        layout = self.layout
        generate_ops(context, context.active_object, layout, draw_label=False)


class HTOOL_PT_VIEW3D_PT_Hair_Panel_Profile(HT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel_Profile"
    bl_label = "Profile"
    bl_parent_id = "HTOOL_PT_VIEW3D_PT_Hair_Panel"
    bl_order = 20

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and (obj.type == 'MESH' and get_profile_mod(obj)) or obj.type == 'CURVES'

    def draw(self, context):
        layout = self.layout
        profile_ops(context, layout, draw_label=False)

# for material_uv_opso


class HTOOL_PT_VIEW3D_PT_Hair_Panel_Material_UV_Ops(HT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel_Material_UV_Ops"
    bl_label = "Material / UV"
    bl_parent_id = "HTOOL_PT_VIEW3D_PT_Hair_Panel"
    bl_order = 30

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'CURVES'

    def draw(self, context):
        layout = self.layout
        material_uv_ops(context.active_object, layout, draw_label=False)


class HTOOL_PT_VIEW3D_PT_Hair_Panel_Rad_Curve(HT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel_Rad_Curve"
    bl_label = "Radius"
    bl_parent_id = "HTOOL_PT_VIEW3D_PT_Hair_Panel"
    bl_order = 40

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'CURVES'

    def draw(self, context):
        layout = self.layout
        rad_curve_box(context, context.active_object, layout, draw_label=False)

# for tilt_curve_box


class HTOOL_PT_VIEW3D_PT_Hair_Panel_Tilt_Curve(HT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel_Tilt_Curve"
    bl_label = "Tilt"
    bl_parent_id = "HTOOL_PT_VIEW3D_PT_Hair_Panel"
    bl_order = 50

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'CURVES'

    def draw(self, context):
        layout = self.layout
        tilt_curve_box(context, context.active_object, layout, draw_label=False)

# for utils_curve_box


class HTOOL_PT_VIEW3D_PT_Hair_Panel_Utils_Curve(HT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel_Utils_Curve"
    bl_label = "Utils"
    bl_parent_id = "HTOOL_PT_VIEW3D_PT_Hair_Panel"
    bl_order = 60

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'CURVES'

    def draw(self, context):
        layout = self.layout
        utils_curve_box(context, context.active_object, layout, draw_label=False)


# for other_curve_box
class HTOOL_PT_VIEW3D_PT_Hair_Panel_Other_Curve(HT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel_Other_Curve"
    bl_label = "Other"
    bl_parent_id = "HTOOL_PT_VIEW3D_PT_Hair_Panel"
    bl_order = 70

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'CURVES'

    def draw(self, context):
        layout = self.layout
        other_curve_box(context, context.active_object, layout, draw_label=False)


# for material_uv_ops
class HTOOL_PT_VIEW3D_PT_Hair_Panel_Material_UV(HT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel_Material_UV"
    bl_label = "Material / UV"
    bl_parent_id = "HTOOL_PT_VIEW3D_PT_Hair_Panel"
    bl_order = 80

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.ht_props.hair_nodes.is_using_gnodes

    def draw(self, context):
        layout = self.layout
        material_uv_ops(context.active_object, layout, draw_label=False)


# for mesh_ops
class HTOOL_PT_VIEW3D_PT_Hair_Panel_Mesh(HT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "HTOOL_PT_VIEW3D_PT_Hair_Panel_Mesh"
    bl_label = "Mesh"
    bl_parent_id = "HTOOL_PT_VIEW3D_PT_Hair_Panel"
    bl_order = 90

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj.type == 'MESH' and context.mode == 'OBJECT' and \
            (obj.ht_props.hair_settings.hair_mesh_parent or obj.ht_props.short_hair_props.hair_base_mesh
             or obj.ht_props.hair_nodes.hair_base_mesh
             or obj.ht_props.hair_nodes.is_convert_result)

    def draw(self, context):
        layout = self.layout
        mesh_ops(context, context.active_object, layout, draw_label=False)


# FOR CURVE EDIT MODE
class HTOOL_PT_VIEW3D_EDIT_Hair_Panel(bpy.types.Panel):
    bl_label = "Hair Tool"
    bl_idname = "HTOOL_PT_VIEW3D_EDIT_Hair_Panel"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Hair Tool'
    bl_context = "curve_edit"

    def draw(self, context):
        HTOOL_PT_VIEW3D_PT_Hair_Panel.draw(self, context)


# create panels that will draw all modifier (geometry nodes based) inputs as props
class ModifierPopupPanel:
    bl_label = "Hair Tool"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Hair Tool'

    modifier_name = ''

    def draw(self, context):
        layout = self.layout
        obj = context.active_object
        modifier = obj.modifiers.get(self.modifier_name)
        if not modifier:
            return
        if modifier.type != 'NODES':
            return

        if bpy.app.version < (4, 0, 0):
            if modifier.node_group:
                for input in modifier.node_group.inputs:
                    if input.type not in {'GEOMETRY'}:
                        input_identifier = input.identifier  # e.g. 'Input_6'
                        layout.prop(modifier, f'["{input_identifier}"]', text=input.name, toggle=input.type == 'BOOLEAN')  # '["Input_6"]'
        else:  # for blender 4 use
            if modifier.node_group:
                for item in modifier.node_group.interface.items_tree:
                    if item.item_type == 'SOCKET' and item.in_out == 'INPUT' and item.socket_type not in {'NodeSocketGeometry'}:
                        input_identifier = item.identifier  # e.g. 'Input_6'
                        layout.prop(modifier, f'["{input_identifier}"]', text=item.name, toggle=item.socket_type == 'NodeSocketBool')  # '["Input_6"]'


# FOR HOTKEY
class HTOOL_MT_HairToolMenu(bpy.types.Menu):
    bl_idname = "HTOOL_MT_HairToolMenu"
    bl_label = "Hair Tool Menu Panel"

    @classmethod
    def poll(cls, context):
        return context.mode in {'EDIT_MESH', 'OBJECT', 'EDIT', 'PAINT_VERTEX', 'PAINT_WEIGHT', 'PARTICLE', 'EDIT_CURVES', 'SCULPT_CURVES'}  # 'EDIT_CURVE',

    def draw(self, context):
        layout = self.layout
        obj = context.active_object

        if context.mode in {'EDIT_CURVES', 'SCULPT_CURVES'}:
            pie = layout.menu_pie()

            # 1 - LEFT
            box = pie.box()
            if context.mode == 'EDIT_CURVES':
                col = box.column(align=True)
                op = col.operator("hair_system.set_uv_regions_new")
                op.force_use_reg_deformer = False

            # 2 - RIGHT
            box = pie.box()

            # 3 - Bottom
            box = pie.split()
            b = box.box()
            rad_curve_box(context, obj, b)
            b = box.box()
            tilt_curve_box(context, obj, b)
            b = box.box()
            utils_curve_box(context, obj, b)
            b = box.box()
            other_curve_box(context, obj, b)

            # 4 - TOP
            box = pie.box()
            profile_ops(context, box)

            # 5 - TOP - LEFT
            if context.mode == 'SCULPT_CURVES':
                box = pie.box()
                attributes_ops(context, box)
            else:
                pie.separator()

        else:  # OBJECT MODE
            pie = layout.menu_pie()

            if obj:
                # 1 - LEFT
                if obj.type == 'CURVES' or obj.ht_props.hair_nodes.is_using_gnodes:
                    box = pie.box()
                    material_uv_ops(obj, box)
                elif obj.ht_props.short_hair_props.is_short_hair:
                    box = pie.box()
                    box.operator('short_hair.import_default_material')  # , icon='IMPORT'
                else:
                    pie.separator()

                # 2 - RIGHT
                box = pie.box()
                generate_ops(context, obj, box)

                # 3 - BOTTOM
                if obj.type == 'MESH' and context.mode == 'OBJECT':
                    if (obj.ht_props.hair_settings.hair_mesh_parent
                         or obj.ht_props.short_hair_props.hair_base_mesh
                         or obj.ht_props.hair_nodes.hair_base_mesh
                         or obj.ht_props.hair_nodes.is_convert_result):  # for ribbon mesh
                        box = pie.box()
                        mesh_ops(context, obj, box)
                    else:
                        pie.separator()
                elif obj.type == 'CURVES' and len(obj.data.curves):
                    box = pie.split()
                    b = box.box()
                    rad_curve_box(context, obj, b)
                    b = box.box()
                    tilt_curve_box(context, obj, b)
                    b = box.box()
                    utils_curve_box(context, obj, b)
                    b = box.box()
                    other_curve_box(context, obj, b)
                elif obj.ht_props.hair_nodes.is_using_gnodes:
                    box = pie.box()
                    box.operator("hsystem.add_ao_mod")  # , icon="VPAINT_HLT"
                else:
                    pie.separator()

                # 4 - TOP
                box = pie.split()

                if (obj.type == 'MESH' and obj.ht_props.hair_nodes.is_using_gnodes) or obj.type == 'CURVES':
                    b = box.box()
                    profile_ops(context, b)

                b = box.box()
                convert_ops(context, obj, b)

                # 5 - TOP - LEFT
                pie.separator()

                # 6 - TOP - RIGHT
                pie.separator()

                # 7 - BOTTOM - LEFT
                pie.separator()

                # 8 - BOTTOM - RIGHT
                pie.separator()

    def check(self, context):
        return True


##########################         Addon Prefernedces                ###############################################
panels = (
    HTOOL_PT_DrawingCurvesPanel,
    HTOOL_PT_Hair_Panel_Bake,
    # HTOOL_PT_HairSystemPanel,
    HTOOL_PT_VIEW3D_PT_Hair_Panel,
    HTOOL_PT_VIEW3D_EDIT_Hair_Panel
)


def update_panel(self, context):
    message = "Hair Tool: Updating Panel locations has failed"
    try:
        for panel in panels:
            if "bl_rna" in panel.__dict__:
                bpy.utils.unregister_class(panel)

        for panel in panels:
            panel.bl_category = get_addon_preferences().category
            bpy.utils.register_class(panel)

    except Exception as e:
        print("\n[{}]\n{}\n\nError:\n{}".format(__name__, message, e))
        pass


addon_keymaps = []  # list of (km_add, kmi_add, descr)


class HtLibDirectories(bpy.types.PropertyGroup):
    def set_abs_path(self, context):
        print(f'Library directory is being changed to absolute. Current: {self.directory}')
        abs_p = os.path.abspath(self.directory)
        if os.path.exists(abs_p):
            self['directory'] = os.path.abspath(self.directory)
        else:
            self['directory'] = str(Path.home())
        print(f"Updated to: {self['directory']}")

    name: bpy.props.StringProperty(name='Library Name', default='Library')
    directory: bpy.props.StringProperty(name="Library directory", default=str(Path.home()), description="Library directory", subtype="DIR_PATH", update=set_abs_path)


class HTLIB_UL_LibDirs(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        path = item
        if path:
            layout.prop(path, 'name', text='', emboss=False)
        else:
            layout.label(text="", translate=False, icon_value=icon)


# operator 'ht_preferences.lib_directory_add'  that will add new item to get_addon_preferences().lib_paths
class HTOOL_OT_LibDirectoryAdd (bpy.types.Operator):
    bl_idname = "ht_preferences.lib_directory_add"
    bl_label = "Add Library Directory"
    bl_description = "Add new library directory"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        prefs = get_addon_preferences()
        prefs.lib_paths.add()
        prefs.lib_active_index = len(prefs.lib_paths) - 1
        self.report({'INFO'}, f"Library directory added")
        return {'FINISHED'}


class HTOOL_OT_LibDirectoryRemove (bpy.types.Operator):
    bl_idname = "ht_preferences.lib_directory_remove"
    bl_label = "Remove Library Directory"
    bl_description = "Remove library directory"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        prefs = get_addon_preferences()
        prefs.lib_paths.remove(prefs.lib_active_index)
        prefs.lib_active_index -= 1
        self.report({'INFO'}, f"Library directory removed")
        return {'FINISHED'}


class hairToolPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    def up_color(self, context):
        # hair_geometry_nodes.FORCE_UPDATE = True # wont work..
        update_hairs_pos(context.active_object, self.draw_color)

    # expandCurveOper: bpy.props.BoolProperty( name="", description="",  default=False)
    # expandGenerators: bpy.props.BoolProperty( name="", description="",  default=False)
    draw_grid_as_wire: bpy.props.BoolProperty(name='Draw grid as wire', description='When generating hair from grid surface, switch mesh display mode to "WIRE"', default=True)
    tabs: bpy.props.EnumProperty(name="Tabs", items=[
                                                     ("OPTIONS", "Options", ""),
                                                     ("BAKING", "Baking", ""),
                                                     ("HOTKEYS", "Hotkeys", ""),
                                                     ("LIBRARY", "Library", "")], default="OPTIONS")
    update_exist: bpy.props.BoolProperty(name="Update Exist", description="There is new GroupPro update",  default=False)
    update_text: bpy.props.StringProperty(name="Update text",  default='')

    category: bpy.props.StringProperty(name="Tab Category", description="Choose a name for the category of the panel", default="Hair Tool", update=update_panel)
    view_render_object_sync: bpy.props.BoolProperty(name="Synchronize outliner  'Eye' and 'Camera' icon",
                                                    description="Synchronize 'Hide View' and 'Hide Render' outliner options for all objects and collections", default=True)

    # * library tab
    lib_paths: bpy.props.CollectionProperty(type=HtLibDirectories)
    lib_active_index: bpy.props.IntProperty(name="Active Index", default=0)

    rerender: bpy.props.BoolProperty(default=False)

    # * baking
    ao_suffix: bpy.props.StringProperty(name='AO', description="Ambient Occlusion", default='_ao')
    diffuse_suffix: bpy.props.StringProperty(name='Diffuse', description="Diffuse", default='_diff')
    opacity_suffix: bpy.props.StringProperty(name='Opacity', description="Opacity", default='_opacity')
    normal_suffix: bpy.props.StringProperty(name='Normal', description="Normal map", default='_normal')
    root_suffix: bpy.props.StringProperty(name='Root', description="Root map", default='_root')
    random_parent: bpy.props.StringProperty(name='Random', description="Random Parent", default='_clump')
    id_suffix: bpy.props.StringProperty(name='Id', description="Random ID mask", default='_id')
    flow_suffix: bpy.props.StringProperty(name='Flow', description="Flow map", default='_flow')
    directional_suffix: bpy.props.StringProperty(name='Directional', description="Directional map", default='_dir')
    depth_suffix: bpy.props.StringProperty(name='Depth', description="Depth map", default='_depth')

    # container_prefix_suffix: bpy.props.EnumProperty(name='', items=[('PREFIX', 'Prefix', ''), ('SUFFIX', 'Suffix', '')], default='PREFIX')

    # Hair Curves - draw settings
    draw_offset: bpy.props.FloatProperty(name="Draw Offset", description="Parent Strand Draw Offset (similar to Draw In Front option in Blender)", min=0.0, soft_max=1.0, default=0.5, subtype='FACTOR')
    draw_color: bpy.props.FloatVectorProperty(name="Color", description="Color", default=(0.8, 0.4, 0.0, 1), subtype='COLOR', size=4, max=1, min=0, update=up_color)

    def draw_key_item(self, layout, description, kc, km, oper_bl_name, prop_name=None, prop_value=None):
        kmi = get_hotkey_entry_item(km, oper_bl_name, prop_name, prop_value)
        if kmi:
            layout.context_pointer_set("keymap", km)
            rna_keymap_ui.draw_kmi([], kc, km, kmi, layout, 0)
        else:
            layout.label(text=description)

    def draw(self, context):
        layout = self.layout
        row = layout.row(align=True)
        # if not CUDA_ACTIVE:
        #     layout.operator('wm.install_pip_lib')
        # else:
        #     layout.operator('wm.uninstall_pip_lib')

        row.prop(self, "tabs", expand=True)
        box = layout.box()
        if self.tabs == "OPTIONS":
            col = box.column()
            col.prop(self, 'draw_grid_as_wire')
            # col.label(text="Tab Category:")
            # col.prop(self, "category", text="")

        elif self.tabs == "BAKING":
            col = box.column()
            sub_col = col.column()
            sub_col.label(text="View vs Render synchronization")
            sub_col.prop(self, 'view_render_object_sync', icon='HIDE_OFF')
            sub_col = col.column()
            sub_col.label(text='Define default suffixes added to baked texture names:')
            sub_col.prop(self, 'ao_suffix')
            sub_col.prop(self, 'diffuse_suffix')
            sub_col.prop(self, 'opacity_suffix')
            sub_col.prop(self, 'normal_suffix')
            sub_col.prop(self, 'root_suffix')
            sub_col.prop(self, 'random_parent')
            sub_col.prop(self, 'id_suffix')
            sub_col.prop(self, 'flow_suffix')
            sub_col.prop(self, 'directional_suffix')
            sub_col.prop(self, 'depth_suffix')

        elif self.tabs == "LIBRARY":
            col = box.column()
            lib_paths = self.lib_paths

            if len(lib_paths) == 0:
                layout.operator("ht_preferences.lib_directory_add", text="Add", icon='ADD')
                return

            row = box.row()
            row.template_list("HTLIB_UL_LibDirs", "", self, "lib_paths", self, "lib_active_index", rows=3)
            col = row.column(align=True)
            col.operator("ht_preferences.lib_directory_add", icon="ADD", text='')
            col.operator("ht_preferences.lib_directory_remove", icon="REMOVE", text='')

            path = lib_paths[self.lib_active_index]
            row = box.row()
            row.prop(path, 'directory', text='')

            sub_row = box.row(align=True)
            sub_row.operator("hair_lib.render_previews", text="Render missing previews") # does it even work?
            sub_row.prop(self, 'rerender', text='Re-render ALL previews')

        elif self.tabs == "HOTKEYS":

            def compare_kmi_props(addon_kmi, user_kmi):
                for prop in dir(addon_kmi.properties):
                    if not prop.startswith('__'):
                        if getattr(addon_kmi.properties, prop) != getattr(user_kmi.properties, prop):
                            return False
                return True

            col = box.column()
            wm = context.window_manager
            old_km_name = ""
            found_addon_kmi = []  # list of tuples (exist, (km, kmi))
            for i, (km_add, kmi_add, desc) in enumerate(addon_keymaps):
                km_user = wm.keyconfigs.user.keymaps.get(km_add.name)  # get user keymap - eg. '3D View', from addon keymap
                # for km_user in wm.keyconfigs.user.keymaps:  # find match in user keymaps
                #     if km_add.name == km_user.name:
                #         km = km_user
                #         break
                got_kmi = False
                for kmi in km_user.keymap_items:
                    if kmi_add.idname == kmi.idname and kmi_add.name == kmi.name and compare_kmi_props(kmi_add, kmi):
                        found_addon_kmi.append((True, (km_user, kmi, desc)))
                        got_kmi = True
                if not got_kmi:
                    found_addon_kmi.append((False, (km_add, kmi_add, desc)))

            # found_addon_kmi = sorted(set(found_addon_kmi), key=found_addon_kmi.index)

            kc = wm.keyconfigs.user
            for i, (exist, (km, kmi, desc)) in enumerate(found_addon_kmi):
                if not km.name == old_km_name:
                    col.label(text=str(km.name), icon="DOT")
                if exist:   # user keymap
                    col.context_pointer_set("keymap", km)
                    rna_keymap_ui.draw_kmi([], kc, km, kmi, col, 0)
                    col.separator()
                else:  # else holds addon keymap - that is missing in user keymap
                    split = col.split(factor=0.1)
                    split.label(text='   ')
                    op = split.operator(HT_OT_ReregisterUserKeym.bl_idname, text=f"Restore: {desc}", icon="DISCLOSURE_TRI_RIGHT")
                    op.index = i
                    col.separator()

                old_km_name = km.name

            # description = "No hotkey entry found for Curves Adjust Radius [Alt+S - default]!"
            # self.draw_key_item(col, description, kc, km, 'curves.editmode_adjust_tilt_rad','mode','RADIUS')

            # XXX: removing  addon keymapw is not working anyway. Only user keymap works. But not stored in addon_keymaps
            # col.operator(HTOOL_OT_HairTool_Clear_Hotkeys.bl_idname, text="Clear hotkeys", icon='REMOVE')
            col.operator(HTOOL_OT_HairTool_Reset_Hotkeys.bl_idname, icon='ADD')
#### end  MOve TO pref file ####################################


def get_hotkey_entry_item(km, kmi_name, prop_key=None, prop_val=None):
    '''
    returns hotkey of specific type, with specific properties.name (keymap is not a dict, so referencing by keys is not enough
    if there are multiple hotkeys!)
    '''
    for i, km_item in enumerate(km.keymap_items):
        if km.keymap_items.keys()[i] == kmi_name:
            if prop_key:  # usually for wm.call_menu_pie
                if getattr(km.keymap_items[i].properties, prop_key) == prop_val:  # eg. 'mode' == 'TILT'
                    return km_item  # TODO: maybe this should return multiple keymaps...
            else:  # for operators
                return km_item

    return None


# def register_keymap(kc=None):
#     global addon_keymaps
#     # XXX: this supposed to be addon keymap but it doesn't work... so whatever
#     if not kc:
#         kc = bpy.context.window_manager.keyconfigs.addon  # only for init register - not always shown in user kc!?
#         write_to_addon_keymaps = True              # write only addon km
#     else:
#         write_to_addon_keymaps = False
#     # kc = bpy.context.window_manager.keyconfigs.user  # for everything else - this users sees
#
#     km = kc.keymaps.get('3D View')
#     if not km:
#         km = kc.keymaps.new(name='3D View', space_type='VIEW_3D')
#     kmi = km.keymap_items.new('wm.call_menu_pie', 'H', 'PRESS', ctrl=True, shift=True)
#     kmi.properties.name = "HTOOL_MT_HairToolMenu"
#     kmi.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km, kmi, 'Hair pie Menu [Ctrl + Shift + H - default]'))
#
#     # gpencil_keymap = bpy.context.window_manager.keyconfigs.user.keymaps.get('Grease Pencil')
#     # km5 = kc.keymaps.new(name='3D View', space_type='VIEW_3D')  # ! use existing '3D View' ! or wont work
#     kmi5 = km.keymap_items.new('object.draw_hair_surf', 'LEFTMOUSE', 'PRESS', key_modifier='D')
#     kmi5.active = False  # activate from init - post_load scene...
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km, kmi5, 'Hair Drawing [D - default]'))
#
#     km2 = kc.keymaps.new(name='Curve', space_type='VIEW_3D')
#     kmi2 = km2.keymap_items.new('wm.call_menu_pie', 'H', 'PRESS', ctrl=True, shift=True)
#     kmi2.properties.name = "HTOOL_MT_HairToolMenu"
#     kmi2.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km2, kmi2, 'Hair pie Menu [Ctrl + Shift + H - default]'))
#
#     km2 = kc.keymaps.new(name='Mesh', space_type='VIEW_3D')
#     kmi2 = km2.keymap_items.new('wm.call_menu_pie', 'H', 'PRESS', ctrl=True, shift=True)
#     kmi2.properties.name = "HTOOL_MT_HairToolMenu"
#     kmi2.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km2, kmi2, 'Hair pie Menu [Ctrl + Shift + H - default]'))
#
#     km3 = kc.keymaps.new(name='Vertex Paint', space_type='VIEW_3D')
#     kmi3 = km3.keymap_items.new('wm.call_menu_pie', 'H', 'PRESS', ctrl=True, shift=True)
#     kmi3.properties.name = "HTOOL_MT_HairToolMenu"
#     kmi3.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km3, kmi3, 'Hair pie Menu [Ctrl + Shift + H - default]'))
#
#     km4 = kc.keymaps.new(name='Weight Paint', space_type='VIEW_3D')
#     kmi4 = km4.keymap_items.new('wm.call_menu_pie', 'H', 'PRESS', ctrl=True, shift=True)
#     kmi4.properties.name = "HTOOL_MT_HairToolMenu"
#     kmi4.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km4, kmi4, 'Hair pie Menu [Ctrl + Shift + H - default]'))
#
#     km5 = kc.keymaps.get('Curves')
#     if not km5:
#         km5 = kc.keymaps.new(name='Curves', space_type="EMPTY")
#     kmi5 = km5.keymap_items.new('curves.hide_selection', 'H', 'PRESS', ctrl=False, shift=False)
#     kmi5.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km5, kmi5, 'Curves Hide [H - default]'))
#
#     kmi5 = km5.keymap_items.new('curves.unhide_selection', 'H', 'PRESS', ctrl=False, alt=True)
#     kmi5.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km5, kmi5, 'Curves Unhide [alt+H - default]'))
#
#     kmi6 = km5.keymap_items.new('curves.hair_duplicate', 'D', 'PRESS', shift=True, alt=False)
#     kmi6.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km5, kmi6, 'Curves Duplicate [Shift + D - default]'))
#
#     kmi7 = km5.keymap_items.new('curves.editmode_adjust_tilt_rad', 'S', 'PRESS', shift=False, alt=True)
#     kmi7.properties.mode = "RADIUS"
#     kmi7.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km5, kmi7, 'Curves Adjust Radius [Shift + S - default]'))
#
#     kmi8 = km5.keymap_items.new('curves.editmode_adjust_tilt_rad', 'T', 'PRESS', shift=False, alt=False, ctrl=True)
#     kmi8.properties.mode = "TILT"
#     kmi8.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km5, kmi8, 'Curves Adjust Tilt [Ctrl + T - default]'))
#
#     kmi9 = km5.keymap_items.new('curves.extrude_curve_ending', 'E', 'PRESS')
#     kmi9.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km5, kmi9, 'Extrude Curve [E - default]'))
#
#     kmi10 = km5.keymap_items.new('wm.call_menu', 'W', 'PRESS', ctrl=False, shift=False)
#     kmi10.properties.name = "HTOOL_MT_CurvesContextMenu"
#     kmi10.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km5, kmi10, 'Curves context menu [W - default]'))
#
#     kmi11 = km5.keymap_items.new('curve.hair_radilt_reset', 'T', 'PRESS', shift=False, alt=True)
#     kmi11.properties.mode = "TILT"
#     kmi11.active = True
#     if write_to_addon_keymaps:  # write only addon km
#         addon_keymaps.append((km5, kmi11, 'Curves Reset Tilt [Alt + T - default]'))



def create_keymap_item(kc, km_name, id_name, properties=None, key='', modifiers=None):
    km = kc.keymaps.get(km_name)
    if not km:
        space = 'VIEW_3D' if km_name not in ('Curves','Vertex Paint') else 'EMPTY' # or else 'Curves', or 'Vertex Paint' wont be registered correctly...
        km = kc.keymaps.new(name=km_name, space_type=space)
    kmi = km.keymap_items.new(id_name, key, 'PRESS', **modifiers)
    if properties:
        for k, v in properties.items():
            setattr(kmi.properties, k, v)
    return km, kmi

def register_keymap(kc=None):
    global addon_keymaps
    # XXX: this supposed to be addon keymap but it doesn't work... so whatever
    if not kc:
        kc = bpy.context.window_manager.keyconfigs.addon  # only for init register - not always shown in user kc!?
        write_to_addon_keymaps = True              # write only addon km
    else:
        write_to_addon_keymaps = False
    # kc = bpy.context.window_manager.keyconfigs.user  # for everything else - this users sees

    # Define the keymap items with their properties
    keymap_items = [
        ('3D View', 'wm.call_menu_pie', {'name': "HTOOL_MT_HairToolMenu"}, 'H', {'ctrl': True, 'shift': True}),
        ('3D View', 'object.draw_hair_surf', None, 'LEFTMOUSE', {'key_modifier': 'D'}),
        ('Curve', 'wm.call_menu_pie', {'name': "HTOOL_MT_HairToolMenu"}, 'H', {'ctrl': True, 'shift': True}),
        ('Mesh', 'wm.call_menu_pie', {'name': "HTOOL_MT_HairToolMenu"}, 'H', {'ctrl': True, 'shift': True}),
        ('Vertex Paint', 'wm.call_menu_pie', {'name': "HTOOL_MT_HairToolMenu"}, 'H', {'ctrl': True, 'shift': True}),
        ('Weight Paint', 'wm.call_menu_pie', {'name': "HTOOL_MT_HairToolMenu"}, 'H', {'ctrl': True, 'shift': True}),
        ('Curves', 'curves.hide_selection', None, 'H', {'ctrl': False, 'shift': False}),
        ('Curves', 'curves.unhide_selection', None, 'H', {'ctrl': False, 'alt': True}),
        ('Curves', 'curves.extrude_curve_ending', None, 'E', {}),
        ('Curves', 'wm.call_menu', {'name': "HTOOL_MT_CurvesContextMenu"}, 'W', {'ctrl': False, 'shift': False}),
        ('Curves', 'curve.hair_radilt_reset', {'mode': "TILT"}, 'T', {'shift': False, 'alt': True}),
        ('Curves', 'curve.deform_curve', None, 'D', {'ctrl':True}),
        # ... (add more keymap items as needed)
    ]
    if bpy.app.version < (4, 1, 0):
        keymap_items.extend([ # these were implemented in 4.1
            ('Curves', 'curves.hair_duplicate', None, 'D', {'shift': True, 'alt': False}),
            ('Curves', 'curves.editmode_adjust_tilt_rad', {'mode': "RADIUS"}, 'S', {'shift': False, 'alt': True}),
            ('Curves', 'curves.editmode_adjust_tilt_rad', {'mode': "TILT"}, 'T', {'shift': False, 'alt': False, 'ctrl': True}),
        ])

    # Create keymap items
    for km_name, id_name, properties, key_code, modifiers in keymap_items:
        km, kmi = create_keymap_item(kc, km_name, id_name, properties=properties, key=key_code, modifiers=modifiers)
        kmi.active = True
        if write_to_addon_keymaps:  # write only addon km
            mods = ''
            [mods.join(f'{k} + ') for k, v in modifiers.items() if v]
            addon_keymaps.append((km, kmi, f'{id_name} [{mods+key_code} - default]'))


class HTOOL_OT_HairTool_Reset_Hotkeys(bpy.types.Operator):
    ''' Add hotkey entry '''
    bl_idname = "hairtool.restore_hotkeys"
    bl_label = "Restore Hotkeys"
    bl_options = {'REGISTER', 'INTERNAL'}

    def execute(self, context):
        # unregister_keymap()
        kc = bpy.context.window_manager.keyconfigs.user
        # unregister_keymap()
        register_keymap(kc)  # when restoring - use user kc
        self.report({'INFO'}, "Hotkey added in User Preferences -> Input -> 3D view")
        return {'FINISHED'}


# operator GP_OT_ReregisterUserKeym, that will re-register user keymap from addon keymap item
class HT_OT_ReregisterUserKeym(bpy.types.Operator):
    bl_idname = "hairtool.reregister_user_keym"
    bl_label = "Restore Hair Tool Hotkey"
    bl_description = "Restore missing Hair Tool hotkey entry"
    bl_options = {'REGISTER', 'INTERNAL'}

    index: bpy.props.IntProperty()

    def execute(self, context):
        (km_add, kmi_add, _) = addon_keymaps[self.index]
        wm = bpy.context.window_manager
        user_keyconfig = wm.keyconfigs.user
        km_usr = user_keyconfig.keymaps.get(km_add.name)
        if not km_usr:
            km_usr = user_keyconfig.keymaps.new(name=km_add.name, space_type=km_add.space_type)
        kmi_user = km_usr.keymap_items.new(kmi_add.idname, kmi_add.type, kmi_add.value, ctrl=kmi_add.ctrl, shift=kmi_add.shift, alt=kmi_add.alt, oskey=kmi_add.oskey, key_modifier=kmi_add.key_modifier)

        def write_kmi_props(addon_kmi, user_kmi):
            for k, v in addon_kmi.properties.items():
                setattr(user_kmi.properties, k, getattr(addon_kmi.properties, k))

        write_kmi_props(kmi_add, kmi_user)

        # if hasattr(kmi_add.properties, 'name'):
        #     kmi.properties.name = kmi_add.properties.name
        self.report({'INFO'}, "Restored Group Pro Hotkeys in User Preferences -> Input -> 3D view")
        return {'FINISHED'}


class HTOOL_OT_HairTool_Clear_Hotkeys(bpy.types.Operator):
    ''' Add hotkey entry '''
    bl_idname = "hairtool.clear_hotkey"
    bl_label = "Addon Preferences Example"
    bl_options = {'REGISTER', 'INTERNAL'}

    def execute(self, context):
        unregister_keymap()
        self.report({'INFO'}, "Hotkey removed in User Preferences -> Input ")
        return {'FINISHED'}


def unregister_keymap():  # this only unregisters obj mode hotkey (what about curve mode hotkey)
    global addon_keymaps
    for km, kmi, desc in addon_keymaps:
        if len(km.keymap_items) > 0:
            try:
                km.keymap_items.remove(kmi)
            except Exception as e:
                print(e)
                print('Failed to remove keymap item:'+desc)
    addon_keymaps.clear()
    # kc = bpy.context.window_manager.keyconfigs.addon
    # km_obj = kc.keymaps.get('3D View')
    # if km_obj:
    #     for kmi in reversed(km_obj.keymap_items):
    #         if kmi.idname == 'object.draw_hair_surf':
    #             km_obj.keymap_items.remove(kmi)
    # if km_obj:
    #     for kmi in reversed(km_obj.keymap_items):
    #         if kmi.idname == 'wm.call_menu_pie' and kmi.properties.name == 'HTOOL_MT_HairToolMenu':
    #             km_obj.keymap_items.remove(kmi)
