# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
from .hair_baking.hair_geometry_nodes_selections import get_tips_pts
from .hair_baking.hair_geometry_nodes_shared import get_profile_mod, create_default_uv_owner_uvs, has_valid_hsys_uvs, set_uv_boxes_deformer, prof_mat_input_name, uv_owner_input_name, auto_update_all_curves_uv_regions, assing_uv_boxes_to_splines, hsys_import_default_mat, store_profile_info
from bpy.props import  IntProperty

hair_mat_name ='HT_Default_Material'


# class HTOOL_OT_CurvesUVRefresh(bpy.types.Operator):
#     bl_label = "Material / UV Refresh"
#     bl_idname = "object.curve_uv_refresh"
#     bl_description = "Assign last used Hair Material to selected curve. Also assign last used hair UVs data, if any is available."
#     bl_options = {"REGISTER", "UNDO"}
#
#     def execute(self, context):
#         self.uvCurveRefresh(context.active_object, refresh_everything=False)
#         context.active_object.update_tag()
#         return {"FINISHED"}
#
#
#     @staticmethod
#     def uvCurveRefresh(curve_obj, refresh_everything=False):
#         ''' What it does: if profile has not mat, or uv_data- then load defautlts. Otherwise it is handled in :get_set_hair_profile_mod
#             If refresh_everything: then force to search HairMatAniso, and force UVs gen with boxes..
#         '''
#         # XXX: move to shared - only leave SET_UV_regions? And 'real' curve ub boxesupdate here? Or not...
#         curve_obj.ht_props.is_curve = True # force material proper aniso dir (even on Hsystem)
#         profile_mod = get_profile_mod(curve_obj)
#         if not profile_mod:
#             print('Cannot refresh UVs on Hair System. No profile_mod found for curve_obj:', curve_obj.name)
#             return
#         if curve_obj.ht_props.hair_nodes.is_using_gnodes: # hair system case:
#             # Refresh everything will Redo the uvs (and hair regions)
#             # Basically only check if we do not have to run create_default_hsys_uvs()
#
#             if not profile_mod[prof_mat_input_name] or not profile_mod[uv_owner_input_name]: # split - if not material, vs if
#                 refresh_everything = True
#
#             # HSys has material, has UVData object, - make sure it has UV_Region_Set deformer too.. or not?
#             # if not refresh_everything: # just add SET_UV_regions if not there
#             #     if profile_mod[prof_mat_input_name] and profile_mod[uv_owner_input_name]:
#             #         any_uv_box_set_node, set_uv_reg_deformers = get_uv_reg_set_deformers(curve_obj)
#             #         if not any_uv_box_set_node:
#             #             print("No 'UV Regions Set' node found. Adding default one")
#             #             set_uv_boxes_deformer(curve_obj, 'ALL')
#             #
#             # if refresh_everything: #  should it attach material too?
#             #     if not has_valid_hsys_uvs(curve_obj):
#             #         create_default_uv_owner_uvs(curve_obj)
#             #     # what about mat...
#             #
#             #     any_uv_box_set_node, set_uv_reg_deformers = get_uv_reg_set_deformers(curve_obj)
#             #     if any_uv_box_set_node: # we have one or more UV_Reg_set deformers
#             #         for reg_set_deform in set_uv_reg_deformers:
#             #             boxes_ids_obj = reg_set_deform.inputs['UV Regions'].default_value
#             #             reg_set_deform.inputs['UV Regions'].default_value = None
#             #             if boxes_ids_obj and boxes_ids_obj.users == 0:
#             #                 bpy.data.objects.remove(boxes_ids_obj)
#             #             reg_set_deform.inputs['Use All Regions'].default_value = True
#             #     else:
#             #         set_uv_boxes_deformer(curve_obj, 'ALL')
#
#                 # if not profile_mod[prof_mat_input_name]: # split - if not material, vs if
#                 #     force_profile_mat_update(curve_obj, profile_mod)
#
#             else:
#                 pass
#                 # if not any_uv_box_set_node:
#                     # print("No 'UV Regions Set' node found. Adding default one")
#                     # set_uv_boxes(curve_obj, 'ALL')
#
#             # force update hack
#             profile_mod.node_group = profile_mod.node_group
#             return None
#
#         # #force refresh everything if no material is assigned, or UV box coutn != len(mat.slots)
#         #
#         # # if not profile_mod[uv_owner_input_name]:
#         # if not has_valid_hsys_uvs(curve_obj):
#         #     create_default_uv_owner_uvs(curve_obj)
#         #
#         # if curve_obj.ht_props.use_auto_uv:
#         #     auto_update_all_curves_uv_regions(curve_obj)
#
#         # curve_obj.update_tag()
#         return profile_mod


class HTOOL_OT_ImportDefaultHairMat(bpy.types.Operator):
    bl_idname = "hair.load_default_mat"
    bl_label = "Import Default Hair Material"
    bl_description = "Import and assigns default hair material to selected object"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(self, context):  # make selection only false, if obj mode
        obj = context.active_object
        return obj and (obj.type == 'CURVES' or obj.ht_props.hair_nodes.is_using_gnodes)

    def execute(self, context):
        obj = context.active_object
        profile_mod = get_profile_mod(obj)
        if profile_mod:
            new_mat = hsys_import_default_mat(obj, profile_mod)
            if profile_mod:
                profile_mod[prof_mat_input_name] = new_mat
                obj.update_tag()
                # profile_mod[uv_owner_input_name] = None # XXX : should assign default UVs

                if not has_valid_hsys_uvs(obj): # make extra sure uvs are set
                    create_default_uv_owner_uvs(obj)

                if obj.ht_props.use_auto_uv:
                    auto_update_all_curves_uv_regions(obj)

                if obj.ht_props.hair_nodes.is_using_gnodes: # create uv set deformer too
                    set_uv_boxes_deformer(obj, 'ALL') # basically reset 'Set UV Regions' deformer

                store_profile_info(context, obj, data_name='last_profile_info') # update last profile info

                # HTOOL_OT_CurvesUVRefresh.uvCurveRefresh(obj, refresh_everything=True)
                # refresh viewport
                obj.modifiers.update()
                obj.update_tag()
                context.view_layer.update()
                context.area.tag_redraw()

                return {"FINISHED"}
            else: # is hair_system but nor profile - must be broken...
                bpy.ops.object.clear_hair_systems()

        return {"FINISHED"}



BOX_ITEMS  = []
def uv_box_items(self, context):
    curve_obj = context.active_object
    profile_mod = get_profile_mod(curve_obj)
    uv_owner = profile_mod[uv_owner_input_name]
    box_count = len(uv_owner.data.vertices)
    uvBoxes = len(uv_owner.data.vertices)
    global BOX_ITEMS
    if len(BOX_ITEMS) != uvBoxes:
        BOX_ITEMS = [(str(i), 'UV box:'+str(i), '') for i in range(uvBoxes)]
    return BOX_ITEMS

class HTOOL_OT_SetUvRegion(bpy.types.Operator):
    bl_idname = "hair.set_uv_region"
    bl_label = "Set UV Region"
    bl_description = "Set selected strands UV to specific region by its ID number"
    bl_options = {"REGISTER", "UNDO"}

    uv_regions: bpy.props.EnumProperty(name='UV Region ID', description='Pick UV region that will be assigned to selected strands',
                                   items=uv_box_items,  options={'ENUM_FLAG'})
    seed: IntProperty(name="Seed", description='Randomize how UV regions are assigned per selected strand',  default=0, min=0, max=1000)

    @classmethod
    def poll(self, context):  # make selection only false, if obj mode
        return context.active_object and context.active_object.type == 'CURVES'

    def draw(self, context):
        layout = self.layout
        if len(self.uv_regions) > 1:
            layout.prop(self, 'seed')
        layout.label(text='Select UVs that will be assigned to strands:')
        layout.prop(self, 'uv_regions')

    def invoke(self, context, event):
        global BOX_ITEMS
        BOX_ITEMS.clear()
        curve_obj = context.active_object
        self.uv_regions = set() #deselect by default all mats...
        self.seed = curve_obj.ht_props.uv_seed

        called_from_ht_wspace = context.workspace.tools.from_space_view3d_mode(mode=context.mode).idname == "hair_tool.hair_transform"
        onlySelection = True if curve_obj.mode == 'EDIT' else called_from_ht_wspace
        self.spl_ids, _ , _ = get_tips_pts(context, curve_obj, world_space=False, only_sel=onlySelection)
        return self.execute(context)

    def execute(self, context):
        if not self.uv_regions:
            return {"FINISHED"}
        boxes_ids = [int(box) for box in self.uv_regions]
        curve_obj = context.active_object
        assing_uv_boxes_to_splines(curve_obj, self.spl_ids, boxes_ids, self.seed)
        return {"FINISHED"}


