# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
import time
import numpy as np
import gpu
from gpu_extras.batch import batch_for_shader
from .utils.gpu_helpers import draw_text_line, draw_gradient_box, draw_background_box
import blf
from .short_hair import update_all_uv_new, ensure_short_hair_mat_setup_ok
from .hair_baking.hair_geometry_nodes_shared import get_profile_mod

shader = gpu.shader.from_builtin('UNIFORM_COLOR')
delta = 14


def offset_adjust_true(point_count, offset_tip, offset_root):  # must be same as in OffsetNode
    x = np.linspace(1, 0, point_count)
    p1 = 1 - np.power(1-x, 5*offset_tip+1)
    p2 = np.power(x, 5*offset_root+1)
    return (p1 + p2)/2


def power_offset_points(point_count, offset_tip, offset_root):  # must be same as in OffsetNode
    x = np.linspace(1, 0, point_count)
    p1 = 1 - np.power(1-x, offset_tip+1)
    p2 = np.power(p1, offset_root+1)
    return p2

def draw_uv_edit_boxes_px(self, context):
    region = context.region

    draw_text_lines = []
    # draw_text_lines.append([f"state {self.state}", 'H1'])
    draw_text_lines.append(["Draw rectangular shape with [LMB]", 'H1'])
    draw_text_lines.append(["Drag corner points to adjust box size", 'H1'])
    draw_text_lines.append(["[Shift+DEL] - reset uv", 'p'])
    draw_text_lines.append(["[C] - Color Popup", 'p'])
    draw_text_lines.append(["", 'p'])
    draw_text_lines.append(["", 'p'])
    draw_text_lines.append(["", 'p'])
    draw_text_lines.append(["ENTER - Finish; ESC / RMB - Cancel", 'p'])


    gpu.state.point_size_set(6)
    gpu.state.line_width_set(1.4)
    if len(self.uv_data_cache) > 0:
        coords = []
        corner_points = []
        region = context.region
        pointsToDraw = self.uv_data_cache[:]
        active_col =  (0.4, 1.0, 0.2, 1)
        gray_col =   (0.8, 0.8, 0.8, 0.7)
        for i,uv_point in enumerate(pointsToDraw): #draw non edited boxes
            if self.state == 'edit_box' and i == self.edit_point_index:  # draw not edited uv boxes
                # del pointsToDraw[self.edit_point_index]
                continue
            x1, y1 = region.view2d.view_to_region(uv_point['start_point'][0], uv_point['start_point'][1], clip = False) # UV (0,1 space) to region space
            x2, y2 = region.view2d.view_to_region(uv_point['end_point'][0], uv_point['end_point'][1], clip = False)     # UV (0,1 space) to region space

            # mouse_x, mouse_y = region.view2d.view_to_region(self.mouse_pos[0], self.mouse_pos[1], clip=False)

            coords = [(x1, y1), (x1, y2), (x2, y2), (x2, y1)]
            batch_box_outline = batch_for_shader(shader, 'LINE_LOOP', {"pos": coords})
            shader.bind()
            if i == self.hovered_box_id:
                shader.uniform_float("color", active_col)
                draw_text_lines[3] = ["[C] - Color Popup | [S] set color", 'p']
                draw_text_lines[4] = ["[DEL] - Delete box", 'p']
                draw_text_lines[5] = [f"[Shift+Scroll] - offset root/tip  {uv_point['offset_root']:.1f} : {uv_point['offset_tip']:.1f}", 'p']
                draw_text_lines[6] = [f"[F] - Flip box Y direction: {uv_point['flip_y']}", 'p']
            else:
                shader.uniform_float("color", gray_col)
            batch_box_outline.draw(shader)

            # -- Drawing horizontal lines
            pts_count = 10
            y_pts = power_offset_points(pts_count, uv_point['offset_tip'], uv_point['offset_root'])  # generate 10 pts
            pts_corrected = np.zeros((2*pts_count, 2), 'f')

            # generate lines by generating point paris:
            #line_x: (x1, y_x), (x2, y_x)  +
            #line_x+1:  (x1, y_x+1), (x2, y_x+1)  + [x+2] etc.

            pts_corrected[::2, 0] = x1
            pts_corrected[1::2, 0] = x2
            pts_corrected[:, 1] = np.repeat((y2 - y1)*y_pts + y1, 2)

            batch_h_lines = batch_for_shader(shader, 'LINES', {"pos": pts_corrected})
            batch_h_lines.draw(shader)


            y_center = (y1+y2)/2
            if uv_point.get('color'): # color not used by short_hair,
                box_col = tuple(uv_point['color'][0:3]) # 4d self.uv_data_cache[box_id]['color']
                grad_active = tuple(b+0.2 for b in box_col)
            else:
                box_col = (0.0, 0.0, 0.0)
                grad_active = (0.2, 0.5, 0.2)

            if i == self.hovered_box_id and y2 < self.mouse_pos[1] < y_center:
                col_tip = grad_active  ## (0.4, 1.0, 0.2)
            else:
                col_tip =  box_col

            if i == self.hovered_box_id and y_center < self.mouse_pos[1] < y1:
                col_root = grad_active # (0.4, 1.0, 0.2)
            else:
                col_root = box_col

            remap = lambda x, low, high: (high - low) * x + low
            # opac_root = remap(uv_point['offset_root'], 0.2, 0.5)
            draw_gradient_box(x0=x1, y0=y1, x1=x2, y1=y_center, color_top=col_root + (0.2,), color_bottom=col_root + (0.2,))
            #tip grad
            # opac_tip = remap(uv_point['offset_tip'], 0.2, 0.5)
            draw_gradient_box(x0=x1, y0=y_center, x1=x2, y1=y2, color_top=col_tip + (0.2,), color_bottom=col_tip + (0.2,) )

            #drawing points corner handles
            corner_points = [(x1, y1), (x2, y2)]
            for pt in corner_points:
                batch = batch_for_shader(shader, 'POINTS', {"pos": [pt]})
                shader.bind()
                if abs(pt[0]-self.mouse_pos[0]) < delta and abs(pt[1]-self.mouse_pos[1]) < delta:
                    shader.uniform_float("color", active_col)
                else:
                    shader.uniform_float("color", gray_col)
                batch.draw(shader)

            draw_background_box(width=50, height=20, left_margin=x1, bottom_margin=y1-25, bg_color = (0., 0., 0., 0.8))
            font_id = 0
            blf.color(font_id, 1.0, 1.0, 1.0, 1.0)
            if bpy.app.version < (4, 0, 0):
                blf.size(font_id, 17, 60)
            else: # for blender 4
                blf.size(font_id, 17)
            blf.position(font_id, x1+5, y1-20, 0)
            blf.draw(font_id, 'ID: '+str(i))

    if self.state in ['new_box', 'edit_box']:  # draw currently edited box
        box = self.uv_data_cache[self.edit_point_index]

        x1, y1 = region.view2d.view_to_region(box['start_point'][0], box['start_point'][1], clip = False) # UV (0,1 space) to region space
        start_point = [x1,y1]

        x2, y2 = region.view2d.view_to_region(box['end_point'][0], box['end_point'][1], clip = False)
        end_point = [x2, y2]

        coords = [(start_point[0], start_point[1]), (start_point[0], end_point[1]),
                (end_point[0], end_point[1]), (end_point[0], start_point[1])]
        batch = batch_for_shader(shader, 'LINE_LOOP', {"pos": coords})
        shader.bind()
        shader.uniform_float("color", active_col)
        batch.draw(shader)

        corner_points = [(start_point[0], start_point[1]),  (end_point[0], end_point[1])]
        batch = batch_for_shader(shader, 'POINTS', {"pos": corner_points})
        shader.bind()
        batch.draw(shader)
    draw_text_line(draw_text_lines, width_override=340)
    gpu.state.point_size_set(1)
    gpu.state.line_width_set(1.0)



# XXX: used for short hair only now.?
class HTOOL_OT_ModalUvRangeSelect(bpy.types.Operator):
    """Draw a line with the mouse"""
    bl_idname = "short_hair.uv_area_select"
    bl_label = "Draw UVs for Hair"
    bl_description = "Draw box in UV editor to define hair ribbons UVs"
    bl_options = {"REGISTER", "UNDO"}

    _handle = None
    UVBackup = [] # when operator is cancelled, restore UVs

    states = ['main', 'new_box', 'edit_box', 'adjust_offset']
    state = 'main'
    current_state = None

    edit_end_point = False #true when updating existing uv box
    edit_start_point = False #true when updating existing uv box

    mouse_pos = []  #we need it to pass as self.mouse_pos to draw handler

    def manual_undo_redo(self, active_obj, undo=True, reset=False): #cos bpy.ops.ed.undo() if fuk up
        active_obj.material_slots[0].material.ht_props.hair_uv_points.clear()
        undo_cnt = len(self.UVBackup)
        if undo_cnt == 0:
            return
        if reset: # when we cancel the operator
            uv_backup_step = self.UVBackup[0]
        else: #NOTE: current state is saved in UVBackup[-1] -> previous in UVBackup[-2]
            if undo:
                self.undo_idx = max(self.undo_idx-1, -undo_cnt) # limit undo to  UVBackup[-undo_steps]
            else:
                self.undo_idx = min(self.undo_idx+1, -1)  # limit redo to  UVBackup[-1]
            # txt = 'undo' if undo else 'redo'
            # print(txt,' self.undo_idx', self.undo_idx)
            uv_backup_step = self.UVBackup[self.undo_idx]

        first_mat = active_obj.material_slots[0].material
        for points in uv_backup_step['uv_data']:
            first_mat.ht_props.hair_uv_points.add()
            for p in points.keys():  # copy start and end point
                first_mat["ht_props"]["hair_uv_points"][-1][p] = points[p]

    @staticmethod
    def save_xy_to_uv_box(context, hair_uv_point,box, region):
        hair_uv_point.start_point = box['start_point']
        hair_uv_point.end_point = box['end_point']

    def init_uv_data_cache_with_defaults(self, uv_data):
        uv_data['start_point'] = [0, 1]
        uv_data['end_point'] = [1, 0]
        uv_data['offset_tip'] = 0
        uv_data['offset_root'] = 0
        uv_data['flip_y'] = False

    def uv_box_reset(self, hair_uv_points):
        hair_uv_points.clear()
        hair_uv_points.add()
        hair_uv_points[-1].start_point = [0, 1]
        hair_uv_points[-1].end_point = [1, 0]

        self.uv_data_cache.clear()
        self.uv_data_cache.append({})
        self.init_uv_data_cache_with_defaults(self.uv_data_cache[-1])


    def initialize_uv_points_cache(self,  hair_uv_points):
        ''' points for drawing UI in uv editor. Cos uvrefresh() may remove material.ht_props.uv- data '''
        # self.uv_data_cache.clear()
        self.uv_data_cache = [] # stores each uv box data
        for uv_point in hair_uv_points:  # draw non edited boxes
            self.uv_data_cache.append({})
            self.uv_data_cache[-1]['start_point'] = [uv_point.start_point[0], uv_point.start_point[1]]  # vies space to UV (0,1 space)
            self.uv_data_cache[-1]['end_point'] = [uv_point.end_point[0], uv_point.end_point[1]]
            self.uv_data_cache[-1]['offset_tip'] = uv_point.offset_tip
            self.uv_data_cache[-1]['offset_root'] = uv_point.offset_root
            self.uv_data_cache[-1]['flip_y'] = uv_point.flip_y

    def update_short_hair(self, obj):
        short_hair_props = obj.data.ht_props.short_hair
        if len(short_hair_props.short_hair_vg_uvs) == 1 and self.refresh_short_hair_uv_boxes: # we got only one layer -Background
            # then it is possible we want to update bg layer uv boxes...
            first_mat = obj.data.materials[0]
            all_uv_boxes = {str(i) for i,_ in enumerate(first_mat.ht_props.hair_uv_points)} # assign all uv boxes to background
            bg_layer = short_hair_props.short_hair_vg_uvs[0]
            bg_layer.uv_regions = all_uv_boxes
        update_all_uv_new(obj)


    def do_material_update(self, obj):
        self.update_short_hair(obj)
        first_mat = obj.data.materials[0]
        if first_mat:
            self.store_backup_step(first_mat)

    def finish(self, context):
        if self._handle:
            bpy.types.SpaceImageEditor.draw_handler_remove(self._handle, 'WINDOW')
        obj = context.active_object
        self.update_short_hair(obj)

        self.report({'INFO'}, 'Finished')
        return {'FINISHED'}

    def modal(self, context, event):
        context.area.tag_redraw()
        region = context.region
        active_obj = context.active_object

        first_mat = active_obj.data.materials[0] if len(active_obj.data.materials) else None

        hair_uv_points = first_mat.ht_props.hair_uv_points
        if event.type == 'MOUSEMOVE':
            self.mouse_pos = [event.mouse_region_x, event.mouse_region_y]
            x, y = region.view2d.region_to_view(self.mouse_pos[0], self.mouse_pos[1])  # view space to UV (0,1 space)
            if self.state == 'edit_box' or self.state == 'new_box':
                box = self.uv_data_cache[self.edit_point_index]
                if self.edit_start_point:
                    box['start_point'] = [x, y]
                    if x > box['end_point'][0]:
                        box['end_point'][0] = x
                    if y < box['end_point'][1]:
                        box['end_point'][1] = y

                if self.edit_end_point or self.state == 'new_box':
                    box['end_point'] = [x, y]
                    if x < box['start_point'][0]:
                        box['start_point'][0] = x
                    if y > box['start_point'][1]:
                        box['start_point'][1] = y

        else:
            self.hovered_box_id = -1
            for i,uv_point in enumerate(self.uv_data_cache): #draw non edited boxes
                x1, y1 = region.view2d.view_to_region(uv_point['start_point'][0], uv_point['start_point'][1], clip = False) # UV (0,1 space) to region space
                x2, y2 = region.view2d.view_to_region(uv_point['end_point'][0], uv_point['end_point'][1], clip = False)     # UV (0,1 space) to region space
                # mouse_x, mouse_y = region.view2d.view_to_region(self.mouse_pos[0], self.mouse_pos[1], clip=False)
                if x1 < self.mouse_pos[0] < x2 and y2 < self.mouse_pos[1] < y1:
                    self.hovered_box_id = i
                    break


        if self.state == 'main':
            x, y = region.view2d.region_to_view(self.mouse_pos[0], self.mouse_pos[1])  # view space to UV (0,1 space)
            if event.type == 'LEFTMOUSE' and event.value == 'PRESS':
                if not(0 < self.mouse_pos[0] < region.width) or not (0 < self.mouse_pos[1] < region.height):
                    return self.finish(context)

                self.edit_point_index = -1
                for point_index, points in enumerate(hair_uv_points):
                    uv_x1, uv_y1 = region.view2d.view_to_region(points['start_point'][0], points['start_point'][1], clip = False) # UV (0,1 space) to region space
                    uv_x2, uv_y2 = region.view2d.view_to_region(points['end_point'][0], points['end_point'][1], clip = False)     # UV (0,1 space) to region space

                    if abs(self.mouse_pos[0]-uv_x1) < delta and  abs(self.mouse_pos[1]-uv_y1) < delta:
                        self.edit_start_point = True
                        self.edit_point_index = point_index
                        self.state = 'edit_box'
                        return {'RUNNING_MODAL'}
                    if abs(self.mouse_pos[0]-uv_x2) < delta and  abs(self.mouse_pos[1]-uv_y2) < delta:
                        self.edit_end_point = True
                        self.edit_point_index = point_index
                        self.state = 'edit_box'
                        return {'RUNNING_MODAL'}


                if self.edit_point_index == -1: #if not edtiting points then start new uv box
                    hair_uv_points.add()
                    hair_uv_points[-1].start_point = [x, y]
                    self.uv_data_cache.append({})
                    self.init_uv_data_cache_with_defaults(self.uv_data_cache[-1])
                    self.uv_data_cache[-1]['start_point'] = [x, y]
                    self.uv_data_cache[-1]['end_point'] = [x, y]
                    self.state = 'new_box'
                    return {'RUNNING_MODAL'}

            elif event.ctrl and event.shift and event.type == 'Z' and event.value == 'PRESS':
                self.manual_undo_redo(active_obj, undo=False)  # ? why not just bpy.ops.ed.undo? cos its fuk up..
                first_mat = active_obj.data.materials[0]
                self.initialize_uv_points_cache(first_mat.ht_props.hair_uv_points)

            elif event.ctrl and event.type == 'Z' and event.value == 'PRESS':
                self.manual_undo_redo(active_obj)  # ? why not just bpy.ops.ed.undo? cos its fuk up..
                first_mat = active_obj.data.materials[0]
                self.initialize_uv_points_cache(first_mat.ht_props.hair_uv_points)

            elif event.type == 'DEL' and event.value == 'PRESS' and event.shift == False:  # delete box/boxes
                if len(hair_uv_points) > 0:
                    for i, hair_uv_point in reversed(list(enumerate(hair_uv_points))):
                        startX, startY = hair_uv_point.start_point
                        endX, endY = hair_uv_point.end_point
                        if startX < x < endX and endY < y < startY:
                            hair_uv_points.remove(i)
                            self.uv_data_cache.pop(i)
                            if len(hair_uv_points) == 0:
                                self.uv_box_reset(hair_uv_points)
                            self.do_material_update(active_obj)
                            break

            elif event.type == 'F' and event.value == 'PRESS':  # delete box/boxes
                if len(hair_uv_points) > 0:
                    for i, hair_uv_point in enumerate(hair_uv_points):
                        startX, startY = hair_uv_point.start_point
                        endX, endY = hair_uv_point.end_point
                        if startX < x < endX and endY < y < startY:
                            hair_uv_points[i].flip_y = not hair_uv_points[i].flip_y
                            self.uv_data_cache[i]['flip_y'] = hair_uv_points[i].flip_y
                            self.do_material_update(active_obj)
                            break

            elif event.type == 'DEL' and event.value == 'PRESS' and event.shift:
                if len(hair_uv_points) > 0:
                    self.uv_box_reset(hair_uv_points)
                    self.do_material_update(active_obj)

            elif event.type in {'RIGHTMOUSE', 'ESC'}:
                if self._handle:
                    bpy.types.SpaceImageEditor.draw_handler_remove(self._handle, 'WINDOW')
                self.manual_undo_redo(active_obj, reset = True)  # ? why not just bpy.ops.ed.undo? cos its fuk up..
                self.report({'INFO'}, 'CANCELLED')
                return {'CANCELLED'}

            elif event.type in {'NUMPAD_ENTER', 'RET'} and time.time()-self.startTime > 1:
                return self.finish(context)

        elif self.state == 'new_box':
            if event.type == 'LEFTMOUSE' and event.value == 'RELEASE':  # finish drawing box, and calculate uv Transformation
                box = self.uv_data_cache[self.edit_point_index]
                hair_uv_point = hair_uv_points[self.edit_point_index]
                self.save_xy_to_uv_box(context, hair_uv_point, box, region)
                self.do_material_update(active_obj)
                self.state = 'main'

        elif self.state == 'edit_box':
            if event.type == 'LEFTMOUSE'  and event.value == 'RELEASE':
                box = self.uv_data_cache[self.edit_point_index]
                hair_uv_point = hair_uv_points[self.edit_point_index]
                self.save_xy_to_uv_box(context, hair_uv_point, box, region)
                self.edit_end_point = False
                self.edit_start_point = False
                self.do_material_update(active_obj)
                self.state = 'main'

        if event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            # allow navigation
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}

    def store_backup_step(self, first_mat):
        uv_backup_step = {}
        uv_backup_step['uv_data'] = []
        for point in first_mat.ht_props.hair_uv_points:
            out_dict = {k: v[:] if str(type(v)) == "<class 'IDPropertyArray'>" else v for k, v in point.items()}
            uv_backup_step['uv_data'].append(out_dict)
        if self.undo_idx < -1: # we have some redo steps - remove them
            for _ in range(self.undo_idx,-1,1):
                # print('popping')
                self.UVBackup.pop()
        self.undo_idx = -1  # ptr to current undo step
        self.UVBackup.append(uv_backup_step)

    def invoke(self, context, event):
        self.startTime = time.time() #to prevetn enter finishing action, when using spacebar search
        active_obj = context.active_object

        if active_obj.ht_props.hair_nodes.is_using_gnodes:
            self.report({'ERROR'}, "For hair system use UV Hair System operator!")
            return {'CANCELLED'}

        if active_obj.type == 'MESH' and not active_obj.ht_props.short_hair_props.is_short_hair:
            self.report({'ERROR'}, "Targets does not seem valid (curve or short hair)! Cancelling")
            return {'CANCELLED'}

        if context.area.type == 'IMAGE_EDITOR':
            if active_obj.mode != 'OBJECT':
                bpy.ops.object.mode_set(mode='OBJECT')

            self.hovered_box_id = -1

            self.refresh_short_hair_uv_boxes = False
            mesh_short_hair_props = active_obj.data.ht_props.short_hair if active_obj.type == 'MESH' and hasattr(active_obj.data.ht_props , 'short_hair') else None
            if mesh_short_hair_props and len(mesh_short_hair_props.short_hair_vg_uvs) == 1: # we got only one layer -Background
                # then it is possible we want to update bg layer uv boxes...
                bg_layer = mesh_short_hair_props.short_hair_vg_uvs[0]
                first_mat = ensure_short_hair_mat_setup_ok(active_obj)
                self.refresh_short_hair_uv_boxes = len(bg_layer.uv_regions) == len(first_mat.ht_props.hair_uv_points)

            self.edit_point_index = -1
            self.UVBackup.clear() # manual undo begin backup
            self.undo_idx = -1  # ptr to current undo step
            self.store_backup_step(first_mat)

            self.initialize_uv_points_cache(first_mat.ht_props.hair_uv_points)
            args = (self, context)
            self._handle = bpy.types.SpaceImageEditor.draw_handler_add(draw_uv_edit_boxes_px, args, 'WINDOW', 'POST_PIXEL')
            self.mouse_pos = [event.mouse_region_x, event.mouse_region_y]
            self.report({'INFO'}, 'Draw box shape in UV editor')
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "You must run this operator in UV editor! Cancelling")
            return {'CANCELLED'}


class HT_UVPanel:
    bl_space_type = 'IMAGE_EDITOR'
    bl_region_type = 'UI'
    bl_category = "Hair Tool UV"


class IMAGE_PT_HToolColorPalette(HT_UVPanel, bpy.types.Panel):
    bl_idname = 'IMAGE_PT_HToolColorPalette'
    bl_label = "UVs Regions Color Palette"
    # bl_parent_id = 'IMAGE_PT_HTool'
    bl_order = 20

    @classmethod
    def poll(cls, context):
        return context.active_object and get_profile_mod(context.active_object)

    def draw(self, context):
        layout = self.layout
        scn_ht_props = context.scene.ht_props
        ht_uv_palette = scn_ht_props.ht_uv_palette
        layout.template_ID(scn_ht_props, "ht_uv_palette", new="palette.new")
        if ht_uv_palette and ht_uv_palette.colors:
            if ht_uv_palette.colors.active:
                layout.template_color_picker(ht_uv_palette.colors.active, "color", value_slider=True)
                layout.prop(ht_uv_palette.colors.active, "color", text="")
            layout.label(text="Use 'Hair UV' operator to assign colors to UV regions (with [S] key)")
        else:
            layout.label(text="No active palette set")
        layout.template_palette(scn_ht_props, "ht_uv_palette", color=True)


class IMAGE_PT_HTool(HT_UVPanel, bpy.types.Panel):
    bl_idname = 'IMAGE_PT_HTool'
    bl_label = "Hair UVs"
    bl_order = 10

    @classmethod
    def poll(cls, context):
        return context.active_object and get_profile_mod(context.active_object)

    def draw(self, context):
        layout = self.layout
        # if context.active_object.ht_props.hair_nodes.is_using_gnodes:
        layout.operator("hair_system.uv_define", text='Define UV Regions', icon='UV_ISLANDSEL')  # Select by UV
        layout.operator("hair_system.reset_uvs", icon='FILE_REFRESH')  # Select by UV
        obj_uv_props = context.active_object.ht_props
        col = layout.column(align=True)
        col.prop(obj_uv_props, 'use_auto_uv', toggle=True)
        if obj_uv_props.use_auto_uv:
            col.prop(obj_uv_props, "uv_seed", )
        # else:
        #    layout.operator("short_hair.uv_area_select", text='Define UV Regions', icon='UV_ISLANDSEL')  # Select by UV



def uvGridShow_button(self, context):
    layout = self.layout
    if context.active_object:
        obj = context.active_object
        if obj.ht_props.short_hair_props.is_short_hair:
            layout.operator("short_hair.uv_area_select", text='Hair UV', icon='UV_ISLANDSEL')  # Select by UV
        if obj.ht_props.hair_nodes.is_using_gnodes or obj.type == 'CURVES':
            layout.operator("hair_system.uv_define", text='Hair UV', icon='UV_ISLANDSEL')  # Select by UV


def append_IMAGE_HT_header():
    bpy.types.IMAGE_HT_header.append(uvGridShow_button)  # inject button into image editor header

def remove_IMAGE_HT_header():
    bpy.types.IMAGE_HT_header.remove(uvGridShow_button)
